#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import datetime
import re
import unittest
import math
from decimal import Decimal
from xml.etree import ElementTree
from elementpath.helpers import LazyPattern, days_from_common_era, \
    months2days, round_number, is_idrefs, collapse_white_spaces, escape_json_string, \
    get_double, numeric_equal, numeric_not_equal, equal, not_equal, \
    match_wildcard, unescape_json_string, split_function_test
from elementpath.xpath30.xpath30_helpers import decimal_to_string, int_to_roman, \
    int_to_month, int_to_weekday, int_to_words, int_to_alphabetic, week_in_month, \
    to_ordinal_en, to_ordinal_it, format_digits, ordinal_suffix


class HelperFunctionsTest(unittest.TestCase):

    def test_lazy_pattern(self):
        pattern = LazyPattern(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$')
        self.assertIsInstance(pattern, LazyPattern)

        class TestPatterns:
            pattern = LazyPattern(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$')

        self.assertIsInstance(TestPatterns.pattern, re.Pattern)
        self.assertIsNotNone(TestPatterns.pattern.match('foo'))
        self.assertIsNone(TestPatterns.pattern.match('foo:bar'))

    def test_node_is_idref_function(self):
        self.assertTrue(is_idrefs(ElementTree.XML('<A>xyz</A>').text))
        self.assertTrue(is_idrefs(ElementTree.XML('<A>xyz abc</A>').text))
        self.assertFalse(is_idrefs(ElementTree.XML('<A>12345</A>').text))
        self.assertTrue(is_idrefs('alpha'))
        self.assertTrue(is_idrefs('alpha beta'))
        self.assertFalse(is_idrefs('12345'))

    def test_days_from_common_era_function(self):
        days4y = 365 * 3 + 366
        days100y = days4y * 24 + 365 * 4
        days400y = days100y * 4 + 1

        self.assertEqual(days_from_common_era(0), 0)
        self.assertEqual(days_from_common_era(1), 365)
        self.assertEqual(days_from_common_era(3), 365 * 3)
        self.assertEqual(days_from_common_era(4), days4y)
        self.assertEqual(days_from_common_era(100), days100y)
        self.assertEqual(days_from_common_era(200), days100y * 2)
        self.assertEqual(days_from_common_era(300), days100y * 3)
        self.assertEqual(days_from_common_era(400), days400y)
        self.assertEqual(days_from_common_era(800), 2 * days400y)
        self.assertEqual(days_from_common_era(-1), -366)
        self.assertEqual(days_from_common_era(-4), -days4y)
        self.assertEqual(days_from_common_era(-5), -days4y - 366)
        self.assertEqual(days_from_common_era(-100), -days100y - 1)
        self.assertEqual(days_from_common_era(-200), -days100y * 2 - 1)
        self.assertEqual(days_from_common_era(-300), -days100y * 3 - 1)
        self.assertEqual(days_from_common_era(-101), -days100y - 366)
        self.assertEqual(days_from_common_era(-400), -days400y)
        self.assertEqual(days_from_common_era(-401), -days400y - 366)
        self.assertEqual(days_from_common_era(-800), -days400y * 2)

    def test_months2days_function(self):
        self.assertEqual(months2days(-119, 1, 12 * 319), 116512)
        self.assertEqual(months2days(200, 1, -12 * 320) - 1, -116877 - 2)

        # 0000 BCE tests
        self.assertEqual(months2days(0, 1, 12), 366)
        self.assertEqual(months2days(0, 1, -12), -365)
        self.assertEqual(months2days(1, 1, 12), 365)
        self.assertEqual(months2days(1, 1, -12), -366)

        # xs:duration ordering related tests
        self.assertEqual(months2days(year=1696, month=9, months_delta=0), 0)
        self.assertEqual(months2days(1696, 9, 1), 30)
        self.assertEqual(months2days(1696, 9, 2), 61)
        self.assertEqual(months2days(1696, 9, 3), 91)
        self.assertEqual(months2days(1696, 9, 4), 122)
        self.assertEqual(months2days(1696, 9, 5), 153)
        self.assertEqual(months2days(1696, 9, 12), 365)
        self.assertEqual(months2days(1696, 9, -1), -31)
        self.assertEqual(months2days(1696, 9, -2), -62)
        self.assertEqual(months2days(1696, 9, -12), -366)

        self.assertEqual(months2days(1697, 2, 0), 0)
        self.assertEqual(months2days(1697, 2, 1), 28)
        self.assertEqual(months2days(1697, 2, 12), 365)
        self.assertEqual(months2days(1697, 2, -1), -31)
        self.assertEqual(months2days(1697, 2, -2), -62)
        self.assertEqual(months2days(1697, 2, -3), -92)
        self.assertEqual(months2days(1697, 2, -12), -366)
        self.assertEqual(months2days(1697, 2, -14), -428)
        self.assertEqual(months2days(1697, 2, -15), -458)

        self.assertEqual(months2days(1903, 3, 0), 0)
        self.assertEqual(months2days(1903, 3, 1), 31)
        self.assertEqual(months2days(1903, 3, 2), 61)
        self.assertEqual(months2days(1903, 3, 3), 92)
        self.assertEqual(months2days(1903, 3, 4), 122)
        self.assertEqual(months2days(1903, 3, 11), 366 - 29)
        self.assertEqual(months2days(1903, 3, 12), 366)
        self.assertEqual(months2days(1903, 3, -1), -28)
        self.assertEqual(months2days(1903, 3, -2), -59)
        self.assertEqual(months2days(1903, 3, -3), -90)
        self.assertEqual(months2days(1903, 3, -12), -365)

        self.assertEqual(months2days(1903, 7, 0), 0)
        self.assertEqual(months2days(1903, 7, 1), 31)
        self.assertEqual(months2days(1903, 7, 2), 62)
        self.assertEqual(months2days(1903, 7, 3), 92)
        self.assertEqual(months2days(1903, 7, 6), 184)
        self.assertEqual(months2days(1903, 7, 12), 366)
        self.assertEqual(months2days(1903, 7, -1), -30)
        self.assertEqual(months2days(1903, 7, -2), -61)
        self.assertEqual(months2days(1903, 7, -6), -181)
        self.assertEqual(months2days(1903, 7, -12), -365)

        # Extra tests
        self.assertEqual(months2days(1900, 3, 0), 0)
        self.assertEqual(months2days(1900, 3, 1), 31)
        self.assertEqual(months2days(1900, 3, 24), 730)
        self.assertEqual(months2days(1900, 3, -1), -28)
        self.assertEqual(months2days(1900, 3, -24), -730)

        self.assertEqual(months2days(1000, 4, 0), 0)
        self.assertEqual(months2days(1000, 4, 1), 30)
        self.assertEqual(months2days(1000, 4, 24), 730)
        self.assertEqual(months2days(1000, 4, -1), -31)
        self.assertEqual(months2days(1000, 4, -24), -730)

        self.assertEqual(months2days(2001, 10, -12), -365)
        self.assertEqual(months2days(2000, 10, -12), -366)
        self.assertEqual(months2days(2000, 2, -12), -365)
        self.assertEqual(months2days(2000, 3, -12), -366)

    def test_round_number_function(self):
        self.assertTrue(math.isnan(round_number(float('NaN'))))
        self.assertTrue(math.isinf(round_number(float('INF'))))
        self.assertTrue(math.isinf(round_number(float('-INF'))))
        self.assertEqual(round_number(10.1), 10)
        self.assertEqual(round_number(9.5), 10)
        self.assertEqual(round_number(-10.1), -10)
        self.assertEqual(round_number(-9.5), -9)

    def test_collapse_white_spaces_function(self):
        self.assertEqual(collapse_white_spaces('  ab  c  '), 'ab c')
        self.assertEqual(collapse_white_spaces('  ab\t\nc  '), 'ab c')

    def test_get_double_function(self):
        self.assertEqual(get_double(1), 1.0)
        self.assertEqual(get_double(1.0), 1.0)

        self.assertIs(get_double('NaN'), math.nan)
        self.assertIs(get_double(float('nan')), math.nan)
        self.assertTrue(math.isinf(get_double('INF')))

        self.assertRaises(ValueError, get_double, 'nan')
        self.assertRaises(ValueError, get_double, 'Inf')
        self.assertRaises(ValueError, get_double, 'alfa')

    def test_numeric_equal_function(self):
        self.assertTrue(numeric_equal(1.0, 1))
        self.assertFalse(numeric_equal(1.000001, 1.0))
        self.assertTrue(numeric_equal(1.0000001, 1.0))
        self.assertFalse(numeric_equal(float('nan'), float('nan')))
        self.assertRaises(TypeError, numeric_equal, 'xyz', 1)

    def test_numeric_not_equal_function(self):
        self.assertFalse(numeric_not_equal(1.0, 1))
        self.assertTrue(numeric_not_equal(1.000001, 1.0))
        self.assertFalse(numeric_not_equal(1.0000001, 1.0))
        self.assertTrue(numeric_not_equal(float('nan'), float('nan')))

    def test_equal_function(self):
        self.assertTrue(equal(1.0, 1))
        self.assertFalse(equal(1.000001, 1.0))
        self.assertFalse(equal(1.0000001, 1.0))
        self.assertTrue(equal(float('nan'), float('nan')))
        self.assertTrue(equal('xyz', 'xyz'))

    def test_not_equal_function(self):
        self.assertFalse(not_equal(1.0, 1))
        self.assertTrue(not_equal(1.000001, 1.0))
        self.assertTrue(not_equal(1.0000001, 1.0))
        self.assertFalse(not_equal(float('nan'), float('nan')))
        self.assertFalse(not_equal('xyz', 'xyz'))

    def test_match_wildcard_function(self):
        self.assertTrue(match_wildcard('foo', '*'))
        self.assertTrue(match_wildcard('foo', '{*}*'))

        # Use only universal names in Clark’s notation:
        # ref. http://www.jclark.com/xml/xmlns.htm
        self.assertFalse(match_wildcard('foo', '*:*'))

        self.assertTrue(match_wildcard('foo', '{*}foo'))
        self.assertFalse(match_wildcard('foo', '{*}bar'))
        self.assertTrue(match_wildcard('{ns}foo', '{*}foo'))
        self.assertFalse(match_wildcard('{ns}foo', '{*}bar'))
        self.assertTrue(match_wildcard('{tns}foo', '{tns}*'))
        self.assertFalse(match_wildcard('{tns}foo', '{bar}*'))
        self.assertTrue(match_wildcard('{ns}foo', '{ns}*'))
        self.assertFalse(match_wildcard('{ns}foo', '{ns}foo'))  # is not a wildcard
        self.assertFalse(match_wildcard('{ns}foo', '{ns}bar'))

    def test_escape_json_string_function(self):
        self.assertEqual(escape_json_string("\""), '\\"')
        self.assertEqual(escape_json_string("\""), '\\"')
        self.assertEqual(escape_json_string('\\"', escaped=True), '\\"')
        self.assertEqual(escape_json_string('\\u000A', escaped=True), '\\u000A')

    def test_unescape_json_string_function(self):
        self.assertEqual(unescape_json_string('foo'), 'foo')
        self.assertEqual(unescape_json_string('\\n'), '\n')
        self.assertEqual(unescape_json_string('\\u0031'), '1')
        self.assertEqual(unescape_json_string('\\"'), '"')
        self.assertEqual(unescape_json_string('\\\\'), '\\')
        self.assertEqual(unescape_json_string('\\u000a'), '\n')
        self.assertEqual(unescape_json_string('\\U0000000a'), '\n')
        self.assertEqual(unescape_json_string('-\\r-'), '-\r-')
        self.assertEqual(unescape_json_string("-\\t-"), '-\t-')

    def test_split_function_test_function(self):
        self.assertEqual(
            split_function_test('element(*)'), []
        )
        self.assertEqual(
            split_function_test('function(*)'), ['*']
        )
        self.assertEqual(
            split_function_test('function(item()) as xs:anyAtomicType'),
            ['item()', 'xs:anyAtomicType']
        )
        self.assertEqual(
            split_function_test('function(xs:string) as xs:integer*'),
            ['xs:string', 'xs:integer*']
        )
        self.assertEqual(
            split_function_test('function() as map(xs:string, item())'),
            ['map(xs:string, item())']
        )
        self.assertEqual(
            split_function_test('function(item()*, item()*, item()*) as item()*'),
            ['item()*', 'item()*', 'item()*', 'item()*']
        )


class XPath30HelperFunctionsTest(unittest.TestCase):

    def test_ordinal_suffix(self):
        self.assertEqual(ordinal_suffix(1), 'st')
        self.assertEqual(ordinal_suffix(2), 'nd')
        self.assertEqual(ordinal_suffix(3), 'rd')
        for n in range(4, 21):
            self.assertEqual(ordinal_suffix(n), 'th')

        self.assertEqual(ordinal_suffix(21), 'st')
        self.assertEqual(ordinal_suffix(22), 'nd')
        self.assertEqual(ordinal_suffix(23), 'rd')
        self.assertEqual(ordinal_suffix(24), 'th')

        self.assertEqual(ordinal_suffix(100), 'th')
        self.assertEqual(ordinal_suffix(100000), 'th')

    def test_to_ordinal_en(self):
        self.assertEqual(to_ordinal_en('one'), 'first')
        self.assertEqual(to_ordinal_en('two'), 'second')
        self.assertEqual(to_ordinal_en('three'), 'third')
        self.assertEqual(to_ordinal_en('four'), 'fourth')
        self.assertEqual(to_ordinal_en('five'), 'fifth')
        self.assertEqual(to_ordinal_en('six'), 'sixth')
        self.assertEqual(to_ordinal_en('seven'), 'seventh')
        self.assertEqual(to_ordinal_en('eight'), 'eighth')
        self.assertEqual(to_ordinal_en('nine'), 'ninth')
        self.assertEqual(to_ordinal_en('ten'), 'tenth')
        self.assertEqual(to_ordinal_en('eleven'), 'eleventh')
        self.assertEqual(to_ordinal_en('twelve'), 'twelfth')
        self.assertEqual(to_ordinal_en('thirteen'), 'thirteenth')
        self.assertEqual(to_ordinal_en('fourteen'), 'fourteenth')
        self.assertEqual(to_ordinal_en('fifteen'), 'fifteenth')
        self.assertEqual(to_ordinal_en('sixteen'), 'sixteenth')
        self.assertEqual(to_ordinal_en('seventeen'), 'seventeenth')
        self.assertEqual(to_ordinal_en('eighteen'), 'eighteenth')
        self.assertEqual(to_ordinal_en('nineteen'), 'nineteenth')
        self.assertEqual(to_ordinal_en('twenty'), 'twentieth')
        self.assertEqual(to_ordinal_en('twenty-one'), 'twenty-first')
        self.assertEqual(to_ordinal_en('twenty-two'), 'twenty-second')
        self.assertEqual(to_ordinal_en('twenty-three'), 'twenty-third')
        self.assertEqual(to_ordinal_en('twenty-four'), 'twenty-fourth')
        self.assertEqual(to_ordinal_en('thirty'), 'thirtieth')
        self.assertEqual(to_ordinal_en('thirty-one'), 'thirty-first')
        self.assertEqual(to_ordinal_en('thirty-two'), 'thirty-second')
        self.assertEqual(to_ordinal_en('thirty-three'), 'thirty-third')
        self.assertEqual(to_ordinal_en('thirty-four'), 'thirty-fourth')
        self.assertEqual(to_ordinal_en('forty'), 'fortieth')
        self.assertEqual(to_ordinal_en('fifty'), 'fiftieth')
        self.assertEqual(to_ordinal_en('sixty'), 'sixtieth')
        self.assertEqual(to_ordinal_en('seventy'), 'seventieth')
        self.assertEqual(to_ordinal_en('eighty'), 'eightieth')
        self.assertEqual(to_ordinal_en('ninety'), 'ninetieth')
        self.assertEqual(to_ordinal_en('one-hundred'), 'one-hundredth')
        self.assertEqual(to_ordinal_en('one hundred'), 'one hundredth')
        self.assertEqual(to_ordinal_en('one hundred and two'), 'one hundred and second')
        self.assertEqual(to_ordinal_en('two-hundred'), 'two-hundredth')

    def test_to_ordinal_it(self):
        self.assertEqual(to_ordinal_it('uno'), 'primo')
        self.assertEqual(to_ordinal_it('due'), 'secondo')
        self.assertEqual(to_ordinal_it('tre'), 'terzo')
        self.assertEqual(to_ordinal_it('quattro'), 'quarto')
        self.assertEqual(to_ordinal_it('sedici'), 'sedicesimo')
        self.assertEqual(to_ordinal_it('diciassette'), 'diciassettesimo')

        self.assertEqual(to_ordinal_it(
            'quindici', '%spellout-ordinal-feminine'), 'quindicesima'
        )

    def test_decimal_to_string(self):
        self.assertEqual(decimal_to_string(Decimal('3.14')), '3.14')
        self.assertEqual(decimal_to_string(Decimal('0.00')), '0.00')
        self.assertEqual(decimal_to_string(Decimal('-0.0')), '-0.0')
        self.assertNotEqual(decimal_to_string(-Decimal('0.0')), '-0.0')
        self.assertEqual(decimal_to_string(-Decimal('90.891')), '-90.891')
        self.assertEqual(decimal_to_string(Decimal('009.00')), '9.00')

    def test_int_to_roman(self):
        self.assertEqual(int_to_roman(1900), 'MCM')
        self.assertEqual(int_to_roman(1), 'I')
        self.assertEqual(int_to_roman(2), 'II')
        self.assertEqual(int_to_roman(3), 'III')
        self.assertEqual(int_to_roman(4), 'IV')
        self.assertEqual(int_to_roman(5), 'V')
        self.assertEqual(int_to_roman(6), 'VI')
        self.assertEqual(int_to_roman(7), 'VII')
        self.assertEqual(int_to_roman(8), 'VIII')
        self.assertEqual(int_to_roman(9), 'IX')
        self.assertEqual(int_to_roman(10), 'X')
        self.assertEqual(int_to_roman(11), 'XI')
        self.assertEqual(int_to_roman(12), 'XII')
        self.assertEqual(int_to_roman(13), 'XIII')
        self.assertEqual(int_to_roman(14), 'XIV')
        self.assertEqual(int_to_roman(15), 'XV')
        self.assertEqual(int_to_roman(48), 'XLVIII')
        self.assertEqual(int_to_roman(49), 'XLIX')
        self.assertEqual(int_to_roman(99), 'XCIX')
        self.assertEqual(int_to_roman(100), 'C')
        self.assertEqual(int_to_roman(499), 'CDXCIX')
        self.assertEqual(int_to_roman(1000), 'M')

    def test_int_to_month(self):
        self.assertEqual(int_to_month(9), 'september')
        self.assertEqual(int_to_month(9, 'it'), 'settembre')
        self.assertEqual(int_to_month(9, 'de'), 'september')
        self.assertRaises(KeyError, int_to_month, 13)

    def test_int_to_weekday(self):
        self.assertEqual(int_to_weekday(1), 'monday')
        self.assertEqual(int_to_weekday(1), 'monday')
        self.assertEqual(int_to_weekday(1, 'it'), 'lunedì')
        self.assertRaises(KeyError, int_to_weekday, 8)

    def test_week_in_month(self):
        dt = datetime.datetime(1900, 1, 1)
        self.assertEqual(week_in_month(dt), 1)
        dt = datetime.datetime(1900, 1, 10)
        self.assertEqual(week_in_month(dt), 2)

    def test_int_to_alphabetic(self):
        self.assertEqual(int_to_alphabetic(1), 'a')
        self.assertEqual(int_to_alphabetic(2), 'b')
        self.assertEqual(int_to_alphabetic(3), 'c')
        self.assertEqual(int_to_alphabetic(40), 'an')

    def test_int_to_words(self):
        self.assertEqual(int_to_words(1), 'one')
        self.assertEqual(int_to_words(2), 'two')
        self.assertEqual(int_to_words(3), 'three')
        self.assertEqual(int_to_words(100), 'one hundred')
        self.assertEqual(int_to_words(8754), 'eight thousand seven hundred and fifty-four')

    def test_format_digits(self):
        self.assertEqual(format_digits('0', '#'), '0')
        self.assertEqual(format_digits('123456789', '000,00,00'), '12345,67,89')
        self.assertEqual(format_digits('150000', '##֊000'), '150֊000')

        self.assertEqual(format_digits('5', '٩', '٠١٢٣٤٥٦٧٨٩'), '٥')
        self.assertEqual(format_digits('6', '٩', '٠١٢٣٤٥٦٧٨٩'), '٦')

        kwargs = dict(grouping_separator=',')
        self.assertEqual(format_digits('987654321', '###,##0,00', **kwargs), '9876,543,21')
        self.assertEqual(format_digits('3', '0000', **kwargs), '0003')
        self.assertEqual(format_digits('10', '##', **kwargs), '10')
        self.assertEqual(format_digits('01', '00', **kwargs), '01')
        self.assertEqual(format_digits('123456789', '###,##,00', **kwargs), '12345,67,89')
        self.assertEqual(format_digits('123456789', '####,###,##,0', **kwargs), '123,456,78,9')
        self.assertEqual(format_digits('123456789', '###,##,00', **kwargs), '12345,67,89')
        self.assertEqual(format_digits('123456789', '####,###,##,0', **kwargs), '123,456,78,9')
        self.assertEqual(format_digits('12', '9,999', **kwargs), '0,012')
        self.assertEqual(format_digits('34', '99', **kwargs), '34')

        args = ['12', '0.000', '0123456789', '#', '.']
        self.assertEqual(format_digits(*args), '0.012')

        args = ['4030201', '!!!,!!!,٠٠٠', '٠١٢٣٤٥٦٧٨٩', '!', ',']
        self.assertEqual(format_digits(*args), '٤,٠٣٠,٢٠١')

        args = ['050600', '٠٠٠٠٠٠', '٠١٢٣٤٥٦٧٨٩', '!', ',']
        self.assertEqual(format_digits(*args), '٠٥٠٦٠٠')

        args = ['4030201', '!!!,!!!,٠٠٠', '٠١٢٣٤٥٦٧٨٩', '!', ',']
        self.assertEqual(format_digits(*args), '٤,٠٣٠,٢٠١')

        args = ['26931', '!!!,!!!', '0123456789', '!', ',']
        self.assertEqual(format_digits(*args), '26,931')

        args = ['400', '!!!', '0123456789', '!', ',']
        self.assertEqual(format_digits(*args), '400')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
if __name__ == '__main__':
    import argparse
    import pathlib
    import memray
    import xml.etree.ElementTree as ElementTree

    from elementpath import DocumentNode, ElementNode, \
        CommentNode, ProcessingInstructionNode, TextNode

    def get_element_tree(source):
        return ElementTree.XML(source)

    parser = argparse.ArgumentParser()
    parser.add_argument('--depth', type=int, default=7,
                        help="the depth of the test XML tree (7 for default)")
    parser.add_argument('--children', type=int, default=3,
                        help="the number of children for each element (3 for default)")
    params = parser.parse_args()

    print('*' * 64)
    print("*** Memory usage estimation of XPath node trees using memray ***")
    print('*' * 64)
    print()

    chunk = 'lorem ipsum'
    for k in range(params.depth - 1, 0, -1):
        chunk = f'<a{k} b{k}="k">{chunk}</a{k}>' * params.children
    xml_source = f'<a0>{chunk}</a0>'

    label = f'{params.depth}x{params.children}'

    outdir = pathlib.Path(__file__).parent.joinpath('out/')
    et_file = outdir.joinpath(f'memray-element-tree-{label}.bin')
    nt_file = outdir.joinpath(f'memray-node-tree-{label}.bin')

    if et_file.is_file():
        et_file.unlink()

    with memray.Tracker(et_file, memory_interval_ms=1, follow_fork=True):
        root = get_element_tree(xml_source)

    if nt_file.is_file():
        nt_file.unlink()

    with memray.Tracker(nt_file, follow_fork=True):
        namespaces = None
        position = 1

        def build_element_node() -> ElementNode:
            global position

            node = ElementNode(elem, parent, position)
            position += 1

            position += len(nsmap) if 'xml' in nsmap else len(nsmap) + 1
            position += len(elem.attrib)
            
            if elem.text is not None:
                node.children.append(TextNode(elem.text, node, position))
                position += 1

            return node

        # Common nsmap
        nsmap = {} if namespaces is None else dict(namespaces)

        if hasattr(root, 'parse'):
            root_node = parent = DocumentNode(root, position)
            position += 1

            elem = root.getroot()
            child = build_element_node()
            parent.children.append(child)
            parent = child
        else:
            elem = root
            parent = None
            root_node = parent = build_element_node()

        root_node.tree.namespaces = nsmap
        children = iter(elem)
        iterators = []
        ancestors = []

        while True:
            for elem in children:
                if not callable(elem.tag):
                    child = build_element_node()
                elif elem.tag.__name__ == 'Comment':  # type: ignore[attr-defined]
                    child = CommentNode(elem, parent, position)
                    position += 1
                else:
                    child = ProcessingInstructionNode(elem, parent, position)

                parent.children.append(child)
                if elem.tail is not None:
                    parent.children.append(TextNode(elem.tail, parent, position))
                    position += 1

                if len(elem):
                    ancestors.append(parent)
                    parent = child
                    iterators.append(children)
                    children = iter(elem)
                    break
            else:
                try:
                    children, parent = iterators.pop(), ancestors.pop()
                except IndexError:
                    break

    print(f"Number of elements: {sum(1 for _ in root.iter())}")
    print(f"Number of nodes: {sum(1 for _ in root_node.iter())}")

    element_nodes = list(x for x in root_node.iter() if isinstance(x, ElementNode))
    print(f"Number of element nodes: {len(element_nodes)}")



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
import glob
import fileinput
import os
import re
import pathlib
import platform
from itertools import chain


class PackageTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_dir = os.path.dirname(os.path.abspath(__file__))
        cls.package_dir = os.path.dirname(cls.test_dir)
        cls.source_dir = os.path.join(cls.package_dir, 'elementpath/')
        cls.missing_debug = re.compile(
            r"(\bimport\s+pdb\b|\bpdb\s*\.\s*set_trace\(\s*\)|\bprint\s*\(|\bbreakpoint\s*\()"
        )
        cls.get_version = re.compile(
            r"(?:\bversion|__version__)(?:\s*=\s*)(\'[^\']*\'|\"[^\"]*\")"
        )
        cls.get_python_requires = re.compile(
            r"(?:\brequires-python\s*=\s*)(\'[^\']*\'|\"[^\"]*\")"
        )
        cls.get_classifier_version = re.compile(
            r"(?:'Programming\s+Language\s+::\s+Python\s+::\s+)(3\.\d+)(?:\s*')"
        )
        cls.get_copyright_info = re.compile(
            r"(Copyright\s\(c\),\s(?:\d{4}-)?(\d{4}), SISSA)"
        )

    @unittest.skipIf(platform.system() == 'Windows', 'Skip on Windows platform')
    def test_missing_debug_statements(self):
        message = "\nFound a debug missing statement at line %d of file %r: %r"
        filename = None
        source_files = glob.glob(os.path.join(self.source_dir, '*.py')) + \
            glob.glob(os.path.join(self.source_dir, '*/*.py'))

        for line in fileinput.input(source_files):
            if fileinput.isfirstline():
                filename = os.path.basename(fileinput.filename())
                if filename == 'generate_categories.py':
                    fileinput.nextfile()
                    continue

            lineno = fileinput.filelineno()

            match = self.missing_debug.search(line)
            self.assertIsNone(
                match, message % (lineno, filename, match.group(0) if match else None)
            )

    @unittest.skipIf(platform.system() == 'Windows', 'Skip on Windows platform')
    def test_wrong_copyright_info(self):
        message = "\nFound a wrong copyright info at line %d of file %r: %r"
        source_files = chain(pathlib.Path(self.package_dir).glob('LICENSE'),
                             pathlib.Path(self.source_dir).glob('**/*.py'),
                             pathlib.Path(self.package_dir).glob('tests/*.py'))
        package_dir = pathlib.Path(self.package_dir)
        year = '2026'

        wrong_copyright_count = 0
        for line in fileinput.input(source_files):
            match = self.get_copyright_info.search(line)
            if match and match.groups()[1] != year:
                filepath = pathlib.Path(fileinput.filename()).relative_to(package_dir)
                lineno = fileinput.filelineno()
                print(message % (lineno, str(filepath), match.groups()[0]), end='')
                wrong_copyright_count += 1
                continue
        else:
            self.assertEqual(wrong_copyright_count, 0, msg="Found wrong copyright info")

    def test_version_matching(self):
        message = "\nFound a different version at line %d of file %r: %r (maybe %r)."
        files = [
            os.path.join(self.source_dir, '__init__.py'),
            os.path.join(self.package_dir, 'pyproject.toml'),
        ]
        version = filename = None
        for line in fileinput.input(files):
            if fileinput.isfirstline():
                filename = fileinput.filename()
            lineno = fileinput.filelineno()

            match = self.get_version.search(line)
            if match is not None:
                if version is None:
                    version = match.group(1).strip('\'\"')
                else:
                    self.assertTrue(
                        version == match.group(1).strip('\'\"'),
                        message % (lineno, filename, match.group(1).strip('\'\"'), version)
                    )

    def test_python_requirement(self):
        files = [
            os.path.join(self.package_dir, 'pyproject.toml'),
            os.path.join(self.package_dir, 'pyproject.toml'),
        ]

        min_version = None

        for line in fileinput.input(files):
            if min_version is None:
                match = self.get_python_requires.search(line)
                if match is not None:
                    min_version = match.group(1).strip('\'\"')
                    self.assertTrue(
                        min_version.startswith('>=3.') and min_version[4:].isdigit(),
                        msg="Wrong python_requires directive in pyproject.toml: %s" % min_version
                    )
                    min_version = min_version[2:]
            else:
                match = self.get_classifier_version.search(line)
                if match is not None:
                    python_version = match.group(1)
                    self.assertEqual(python_version[:2], min_version[:2])
                    self.assertGreaterEqual(int(python_version[2:]), int(min_version[2:]))

        self.assertIsNotNone(min_version, msg="Missing python_requires in pyproject.toml")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

def main() -> None:
    from io import StringIO
    from xml.etree import ElementTree
    from elementpath import XPath2Parser, XPathContext, DocumentNode
    from elementpath import XPathToken

    parser = XPath2Parser()
    token = parser.parse('/root/(: comment :) child[@attr]')
    assert isinstance(token, XPathToken)
    assert token.tree == '(/ (/ (root)) ([ (child) (@ (attr))))'
    assert token.source == '/root/child[@attr]'

    root = ElementTree.XML('<root><child/><child attr="10"/></root>')
    context = XPathContext(root)
    value = token.evaluate(context)
    print(value)

    token = parser.parse('concat("foo", " ", "bar")')
    assert context.root is not None and token.evaluate() == 'foo bar'

    doc = ElementTree.parse(StringIO('<root><child1/><child2/><child3/></root>'))
    context = XPathContext(doc)  # error?
    assert isinstance(context.root, DocumentNode)
    assert context.document is context.root
    assert context.item is context.root


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
import warnings
import xml.etree.ElementTree as ElementTree

from elementpath import select, iter_select, Selector, XPath2Parser

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None


class XPathSelectorsTest(unittest.TestCase):
    etree = ElementTree

    @classmethod
    def setUpClass(cls) -> None:
        cls.root = cls.etree.XML('<author>Dickens</author>')

    def test_select_function(self):
        self.assertEqual(select(self.root, 'text()'), ['Dickens'])
        self.assertEqual(select(self.root, '$a', variables={'a': 1}), 1)

        self.assertEqual(
            select(self.root, '$a', variables={'a': 1}, variable_types={'a': 'xs:decimal'}), 1
        )

    def test_iter_select_function(self):
        self.assertEqual(list(iter_select(self.root, 'text()')), ['Dickens'])
        self.assertEqual(list(iter_select(self.root, '$a', variables={'a': True})), [True])

    def test_selector_class(self):
        selector = Selector('/A')
        self.assertEqual(repr(selector), "Selector(path='/A', parser=XPath2Parser)")
        self.assertEqual(selector.namespaces, XPath2Parser.DEFAULT_NAMESPACES)

        selector = Selector('text()')
        self.assertEqual(selector.select(self.root), ['Dickens'])
        self.assertEqual(list(selector.iter_select(self.root)), ['Dickens'])

        selector = Selector('$a')
        self.assertEqual(selector.select(self.root, variables={'a': 1}), 1)
        self.assertEqual(list(selector.iter_select(self.root, variables={'a': 1})), [1])

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            selector = Selector('$a', variables={'a': 1})
            self.assertEqual(len(w), 1)
            self.assertEqual(w[0].category, DeprecationWarning)

        self.assertEqual(selector.select(self.root), 1)
        self.assertEqual(list(selector.iter_select(self.root)), [1])

    def test_issue_001(self):
        selector = Selector("//FullPath[ends-with(., 'Temp')]")
        self.assertEqual(selector.select(self.etree.XML('<A/>')), [])
        self.assertEqual(selector.select(self.etree.XML('<FullPath/>')), [])
        root = self.etree.XML('<FullPath>High Temp</FullPath>')
        self.assertEqual(selector.select(root), [root])

    def test_issue_042(self):
        selector1 = Selector('text()')
        selector2 = Selector('sup[last()]/preceding-sibling::text()')
        root = self.etree.XML('<root>a<sup>1</sup>b<sup>2</sup>c<sup>3</sup></root>')
        self.assertEqual(selector1.select(root), selector2.select(root))

        selector2 = Selector('sup[1]/following-sibling::text()')
        root = self.etree.XML('<root><sup>1</sup>b<sup>2</sup>c<sup>3</sup>d</root>')
        self.assertEqual(selector1.select(root), selector2.select(root))

    def test_fragment_argument__issue_081(self):
        # xml1 contains the xml-stylesheet tag
        xml1 = b"""<?xml version="1.0" encoding="UTF-8"?>
        <?xml-stylesheet type='text/xsl' href='test.xsl'?>
        <root>
            <first>
                <second>
                    value
                </second>
            </first>
        </root>
        """

        # the same as xml1, but without the xml-stylesheet tag
        xml2 = b"""<?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <root>
            <first>
                <second>
                    value
                </second>
            </first>
        </root>
        """

        root1 = self.etree.XML(xml1)
        root2 = self.etree.XML(xml2)
        query = "first/second"

        if hasattr(root1, 'xpath'):
            self.assertEqual(select(root1, query), [])
        else:
            self.assertEqual(select(root1, query), root1[0][:])

        self.assertEqual(select(root1, query, fragment=True), root1[0][:])
        self.assertEqual(select(root2, query), root2[0][:])

        self.assertEqual(select(root1, query, fragment=False), [])
        self.assertEqual(select(root2, query, fragment=False), [])

        query = "root/first/second"

        self.assertEqual(select(root1, query, fragment=False), root1[0][:])
        self.assertEqual(select(root2, query, fragment=False), root2[0][:])


@unittest.skipIf(lxml_etree is None, "The lxml library is not installed")
class LxmlXPathSelectorsTest(XPathSelectorsTest):
    etree = lxml_etree

    def test_issue_058(self):
        tei = """<?xml version='1.0' encoding='UTF8'?>
        <?xml-model type="application/xml"?>
        <TEI xmlns="http://www.tei-c.org/ns/1.0">
          <text>
                <pb n="page1"/>
                <pb n="page2"/>
          </text>
        </TEI>
        """

        doc = self.etree.XML(tei.encode())
        namespaces = {'': "http://www.tei-c.org/ns/1.0"}
        k = None
        for k, p in enumerate(select(doc, '//pb', namespaces), start=1):
            self.assertEqual(p.attrib['n'], f'page{k}')
            self.assertEqual(p.xpath('./@n'), [f'page{k}'])
            self.assertEqual(select(doc, './@n'), [])
            self.assertEqual(select(p, './@n'), [f'page{k}'])
            self.assertEqual(select(doc, './@n', item=p), [f'page{k}'])
        else:
            self.assertEqual(k, 2)

    def test_issue_074(self):
        root = lxml_etree.XML("<root><trunk><branch></branch></trunk></root>")

        result = select(root, "trunk")
        self.assertEqual(result, [root[0]])  # [<Element trunk at 0x...>]

        result = select(root, "/root/trunk")
        self.assertEqual(result, [root[0]])  # [<Element trunk at 0x...>]

        root = lxml_etree.XML("<!--comment--><root><trunk><branch></branch></trunk></root>")

        result = select(root, "trunk")
        self.assertEqual(result, [])

        result = select(root, "root/trunk")
        self.assertEqual(result, [root[0]])  # [<Element trunk at 0x...>]

        result = select(root, "/root/trunk")
        self.assertEqual(result, [root[0]])  # [<Element trunk at 0x...>]


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import sys
import unittest
import platform
import io
from pathlib import Path
from xml.dom import pulldom
from xml.etree import ElementTree

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

from elementpath.etree import SafeExpatParser, defuse_xml, \
    etree_tostring, is_etree_document, is_lxml_etree_element, \
    is_lxml_etree_document, etree_deep_equal, etree_iter_paths
from elementpath.exceptions import XMLResourceForbidden


XML_WITH_NAMESPACES = '<pfa:root xmlns:pfa="http://xpath.test/nsa">\n' \
                      '  <pfb:elem xmlns:pfb="http://xpath.test/nsb"/>\n' \
                      '</pfa:root>'


class TestDefuse(unittest.TestCase):

    def test_safe_expat_parser(self):
        parser = SafeExpatParser()
        resource = io.StringIO(XML_WITH_NAMESPACES)
        nodes = []

        for event, node in pulldom.parse(resource, parser):
            self.assertIn(event, (
                'START_DOCUMENT', 'START_ELEMENT', 'CHARACTERS', 'END_ELEMENT',
            ))
            nodes.append(node)
        self.assertEqual(len(nodes), 8)

    def test_defuse_xml_entities(self):
        xml_file = Path(__file__).parent.joinpath('resources/with_entity.xml')

        elem = ElementTree.parse(str(xml_file)).getroot()
        self.assertEqual(elem.text, 'abc')

        with self.assertRaises(XMLResourceForbidden) as ctx:
            with xml_file.open() as fp:
                defuse_xml(fp.read())
        self.assertEqual("Entities are forbidden (entity_name='e')", str(ctx.exception))

    def test_defuse_xml_external_entities(self):
        xml_file = Path(__file__).parent.joinpath('resources/external_entity.xml')

        with self.assertRaises(ElementTree.ParseError) as ctx:
            ElementTree.parse(str(xml_file))
        self.assertIn("undefined entity &ee", str(ctx.exception))

        with self.assertRaises(XMLResourceForbidden) as ctx:
            with xml_file.open() as fp:
                defuse_xml(fp.read())
        self.assertEqual("Entities are forbidden (entity_name='ee')", str(ctx.exception))

    def test_defuse_xml_unused_external_entities(self):
        xml_file = str(Path(__file__).parent.joinpath('resources/unused_external_entity.xml'))

        elem = ElementTree.parse(xml_file).getroot()
        self.assertEqual(elem.text, 'abc')

        with self.assertRaises(XMLResourceForbidden) as ctx:
            with open(xml_file) as fp:
                defuse_xml(fp.read())
        self.assertEqual("Entities are forbidden (entity_name='ee')", str(ctx.exception))

    def test_defuse_xml_unparsed_entities(self):
        xml_file = Path(__file__).parent.joinpath('resources/unparsed_entity.xml')

        with self.assertRaises(XMLResourceForbidden) as ctx:
            with xml_file.open() as fp:
                defuse_xml(fp.read())
        self.assertEqual("Unparsed entities are forbidden (entity_name='logo_file')",
                         str(ctx.exception))

    def test_defuse_xml_unused_unparsed_entities(self):
        xml_file = Path(__file__).parent.joinpath('resources/unused_unparsed_entity.xml')

        elem = ElementTree.parse(str(xml_file)).getroot()
        self.assertIsNone(elem.text)

        with self.assertRaises(XMLResourceForbidden) as ctx:
            with xml_file.open() as fp:
                defuse_xml(fp.read())
        self.assertEqual("Unparsed entities are forbidden (entity_name='logo_file')",
                         str(ctx.exception))


# noinspection PyTypeChecker
class TestElementTree(unittest.TestCase):

    def test_element_string_serialization(self):
        self.assertRaises(TypeError, etree_tostring, '<element/>')

        elem = ElementTree.Element('element')
        self.assertEqual(etree_tostring(elem), '<element />')
        self.assertEqual(etree_tostring(elem, xml_declaration=True), '<element />')

        self.assertEqual(etree_tostring(elem, encoding='us-ascii'), b'<element />')
        self.assertEqual(etree_tostring(elem, encoding='us-ascii', indent='    '),
                         b'    <element />')
        self.assertEqual(etree_tostring(elem, encoding='us-ascii', xml_declaration=True),
                         b'<?xml version="1.0" encoding="us-ascii"?>\n<element />')

        elem.text = '\t'
        self.assertEqual(etree_tostring(elem), '<element>    </element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=2), '<element>  </element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=0), '<element></element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=None), '<element>\t</element>')

        elem.text = '\n\n'
        self.assertEqual(etree_tostring(elem), '<element>\n\n</element>')
        self.assertEqual(etree_tostring(elem, indent='  '), '  <element>\n\n  </element>')

        elem.text = '\nfoo\n'
        self.assertEqual(etree_tostring(elem), '<element>\nfoo\n</element>')
        self.assertEqual(etree_tostring(elem, indent=' '), ' <element>\n foo\n </element>')

        elem.text = None

        self.assertEqual(etree_tostring(elem, encoding='ascii'),
                         b"<?xml version='1.0' encoding='ascii'?>\n<element />")
        self.assertEqual(etree_tostring(elem, encoding='ascii', xml_declaration=False),
                         b'<element />')
        self.assertEqual(etree_tostring(elem, encoding='utf-8'), b'<element />')
        self.assertEqual(etree_tostring(elem, encoding='utf-8', xml_declaration=True),
                         b'<?xml version="1.0" encoding="utf-8"?>\n<element />')

        self.assertEqual(etree_tostring(elem, encoding='iso-8859-1'),
                         b"<?xml version='1.0' encoding='iso-8859-1'?>\n<element />")
        self.assertEqual(etree_tostring(elem, encoding='iso-8859-1', xml_declaration=False),
                         b"<element />")

        self.assertEqual(etree_tostring(elem, method='html'), '<element></element>')
        self.assertEqual(etree_tostring(elem, method='text'), '')

        root = ElementTree.XML('<root>\n'
                               '  text1\n'
                               '  <elem>text2</elem>\n'
                               '</root>')
        self.assertEqual(etree_tostring(root, method='text'), '\n  text1\n  text2')
        self.assertEqual(etree_tostring(root, max_lines=1), '<root>\n  ...\n  ...\n</root>')

        root = ElementTree.XML(XML_WITH_NAMESPACES)
        result = etree_tostring(root)
        self.assertNotEqual(result, XML_WITH_NAMESPACES)
        self.assertNotIn('pxa', result)
        self.assertNotIn('pxa', result)
        self.assertRegex(result, r'xmlns:ns\d="http://xpath.test/nsa')
        self.assertRegex(result, r'xmlns:ns\d="http://xpath.test/nsb')

        namespaces = {
            'pxa': "http://xpath.test/nsa",
            'pxb': "http://xpath.test/nsb"
        }
        expected = '<pxa:root xmlns:pxa="http://xpath.test/nsa" ' \
                   'xmlns:pxb="http://xpath.test/nsb">\n' \
                   '  <pxb:elem />\n' \
                   '</pxa:root>'
        self.assertEqual(etree_tostring(root, namespaces), expected)

        namespaces = {
            '': "http://xpath.test/nsa",
            'pxa': "http://xpath.test/nsa",
            'pxb': "http://xpath.test/nsb"
        }
        self.assertEqual(etree_tostring(root, namespaces), expected)

        namespaces = {
            '': "http://xpath.test/nsa",
            'pxb': "http://xpath.test/nsb"
        }
        expected = '<root xmlns="http://xpath.test/nsa" ' \
                   'xmlns:pxb="http://xpath.test/nsb">\n' \
                   '  <pxb:elem />\n' \
                   '</root>'
        self.assertEqual(etree_tostring(root, namespaces), expected)

    @unittest.skipIf(lxml_etree is None, 'lxml is not installed ...')
    def test_lxml_element_string_serialization(self):
        elem = lxml_etree.Element('element')
        self.assertEqual(etree_tostring(elem), '<element/>')
        self.assertEqual(etree_tostring(elem, xml_declaration=True), '<element/>')

        self.assertEqual(etree_tostring(elem, encoding='us-ascii'), b'<element/>')
        self.assertEqual(etree_tostring(elem, encoding='us-ascii', indent='    '),
                         b'    <element/>')

        elem.text = '\t'
        self.assertEqual(etree_tostring(elem), '<element>    </element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=2), '<element>  </element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=0), '<element></element>')
        self.assertEqual(etree_tostring(elem, spaces_for_tab=None), '<element>\t</element>')
        elem.text = None

        self.assertEqual(etree_tostring(elem, encoding='us-ascii'), b'<element/>')
        self.assertEqual(etree_tostring(elem, encoding='us-ascii', xml_declaration=True),
                         b'<?xml version="1.0" encoding="us-ascii"?>\n<element/>')

        self.assertEqual(etree_tostring(elem, encoding='ascii'), b'<element/>')
        self.assertEqual(etree_tostring(elem, encoding='ascii', xml_declaration=True),
                         b'<?xml version="1.0" encoding="ascii"?>\n<element/>')

        self.assertEqual(etree_tostring(elem, encoding='utf-8'), b'<element/>')
        self.assertEqual(etree_tostring(elem, encoding='utf-8', xml_declaration=True),
                         b'<?xml version="1.0" encoding="utf-8"?>\n<element/>')

        self.assertEqual(etree_tostring(elem, encoding='iso-8859-1'),
                         b"<?xml version='1.0' encoding='iso-8859-1'?>\n<element/>")
        self.assertEqual(etree_tostring(elem, encoding='iso-8859-1', xml_declaration=False),
                         b"<element/>")

        self.assertEqual(etree_tostring(elem, method='html'), '<element></element>')
        self.assertEqual(etree_tostring(elem, method='text'), '')

        root = lxml_etree.XML('<root>\n'
                              '  text1\n'
                              '  <elem>text2</elem>\n'
                              '</root>')
        self.assertEqual(etree_tostring(root, method='text'), '\n  text1\n  text2')

        root = lxml_etree.XML(XML_WITH_NAMESPACES)
        self.assertEqual(etree_tostring(root), XML_WITH_NAMESPACES)

        namespaces = {
            'tns0': "http://xpath.test/nsa",
            'tns1': "http://xpath.test/nsb"
        }
        self.assertEqual(etree_tostring(root, namespaces), XML_WITH_NAMESPACES)

        for prefix, uri in namespaces.items():
            lxml_etree.register_namespace(prefix, uri)
        self.assertEqual(etree_tostring(root), XML_WITH_NAMESPACES)

    def test_is_etree_document_function(self):
        document = ElementTree.parse(io.StringIO('<A/>'))
        self.assertTrue(is_etree_document(document))
        self.assertFalse(is_etree_document(ElementTree.XML('<A/>')))

    def test_is_lxml_etree_document_function(self):
        document = ElementTree.parse(io.StringIO('<A/>'))
        self.assertFalse(is_lxml_etree_document(document))
        if lxml_etree is not None:
            document = lxml_etree.parse(io.StringIO('<A/>'))
            self.assertTrue(is_lxml_etree_document(document))
            self.assertFalse(is_lxml_etree_document(lxml_etree.XML('<A/>')))

    def test_is_lxml_etree_element_function(self):
        self.assertFalse(is_lxml_etree_element(ElementTree.XML('<A/>')))
        if lxml_etree is not None:
            self.assertTrue(is_lxml_etree_element(lxml_etree.XML('<A/>')))

    def test_etree_deep_equal_function(self):
        e1 = ElementTree.XML('<root a="foo"/>')
        e2 = ElementTree.XML('<root a="foo"/>')
        self.assertTrue(etree_deep_equal(e1, e2))

        e2 = ElementTree.XML('<ROOT a="foo"/>')
        self.assertFalse(etree_deep_equal(e1, e2))

        e2 = ElementTree.XML('<root a="bar"/>')
        self.assertFalse(etree_deep_equal(e1, e2))

        e2 = ElementTree.XML('<root a="foo">bar</root>')
        self.assertFalse(etree_deep_equal(e1, e2))

    def test_etree_iter_paths_function(self):
        root = ElementTree.XML('<root><child/></root>')
        result = list(etree_iter_paths(root))
        self.assertEqual(
            result, [(root, '.'), (root[0], './Q{}child[1]')]
        )

        root = ElementTree.XML('<root><tns:child xmlns:tns="http://xpath.test/ns"/></root>')
        result = list(etree_iter_paths(root))
        self.assertEqual(
            result, [(root, '.'), (root[0], './Q{http://xpath.test/ns}child[1]')]
        )

        if sys.version_info >= (3, 8):
            parser = ElementTree.XMLParser(
                target=ElementTree.TreeBuilder(insert_comments=True)
            )
            root = ElementTree.XML('<root><!-- comment --></root>', parser)
            result = list(etree_iter_paths(root))
            self.assertEqual(
                result, [(root, '.'), (root[0], './comment()[1]')]
            )
            parser = ElementTree.XMLParser(
                target=ElementTree.TreeBuilder(insert_pis=True)
            )
            root = ElementTree.XML(
                '<root><?xml-stylesheet type="text/xsl" href="style.xsl"?></root>', parser
            )
            result = list(etree_iter_paths(root))
            self.assertEqual(
                result, [(root, '.'), (root[0], './processing-instruction(xml-stylesheet)[1]')]
            )

        if lxml_etree is not None:
            root = lxml_etree.XML('<root><!-- comment --></root>')
            result = list(etree_iter_paths(root))
            self.assertEqual(
                result, [(root, '.'), (root[0], './comment()[1]')]
            )

            root = lxml_etree.XML(
                '<root><?xml-stylesheet type="text/xsl" href="style.xsl"?></root>'
            )
            result = list(etree_iter_paths(root))
            self.assertEqual(
                result, [(root, '.'), (root[0], './processing-instruction(xml-stylesheet)[1]')]
            )


if __name__ == '__main__':
    header_template = "ElementTree tests for elementpath with Python {} on {}"
    header = header_template.format(platform.python_version(), platform.platform())
    print('{0}\n{1}\n{0}'.format("*" * len(header), header))

    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2023-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import decimal
import sys
import unittest
from textwrap import dedent
from xml.etree import ElementTree

try:
    import xmlschema
except ImportError:
    xmlschema = None

from elementpath import get_node_tree
from elementpath.datatypes import QName, UntypedAtomic
from elementpath.xpath_nodes import TextNode, CommentNode, \
    ProcessingInstructionNode
from elementpath.xpath_tokens import XPathMap, XPathArray
from elementpath.serialization import get_serialization_params, \
    serialize_to_xml, serialize_to_json
from elementpath.xpath3 import XPath31Parser


class SerializationTest(unittest.TestCase):

    def test_get_serialization_params_from_map(self):
        parser = XPath31Parser()

        params = XPathMap(parser, items={})
        result = get_serialization_params(params)
        self.assertEqual(result, {})

        params = XPathMap(parser, items={1: 'xml', 'method': None})
        result = get_serialization_params(params)
        self.assertEqual(result, {})

        params = XPathMap(parser, items={'method': 'xml'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'method': 'xml'})

        params = XPathMap(parser, items={'method': 'json'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'method': 'json'})

        params = XPathMap(parser, items={'method': 'adaptive'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'method': 'adaptive'})

        params = XPathMap(parser, items={'indent': True})
        result = get_serialization_params(params)
        self.assertEqual(result, {'indent': True})

        params = XPathMap(parser, items={'indent': False})
        result = get_serialization_params(params)
        self.assertEqual(result, {'indent': False})

        params = XPathMap(parser, items={'indent': UntypedAtomic('true')})
        result = get_serialization_params(params)
        self.assertEqual(result, {'indent': True})

        params = XPathMap(parser, items={'indent': UntypedAtomic('false')})
        result = get_serialization_params(params)
        self.assertEqual(result, {'indent': False})

        params = XPathMap(parser, items={'indent': 'yes'})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'method': 'other'})
        with self.assertRaises(ValueError) as ctx:
            get_serialization_params(params)
        self.assertIn('SEPM0017', str(ctx.exception))

        params = XPathMap(parser, items={'use-character-maps': True})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'omit-xml-declaration': True})
        result = get_serialization_params(params)
        self.assertEqual(result, {'xml_declaration': False})

        params = XPathMap(parser, items={'omit-xml-declaration': False})
        result = get_serialization_params(params)
        self.assertEqual(result, {'xml_declaration': True})

        params = XPathMap(parser, items={'omit-xml-declaration': 'no'})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'item-separator': ';'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'item_separator': ';'})

        params = XPathMap(parser, items={'item-separator': UntypedAtomic('  ')})
        result = get_serialization_params(params)
        self.assertEqual(result, {'item_separator': '  '})

        params = XPathMap(parser, items={'item-separator': True})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'encoding': 'ISO-8859-1'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'encoding': 'ISO-8859-1'})

        params = XPathMap(parser, items={'encoding': False})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'allow-duplicate-names': True})
        result = get_serialization_params(params)
        self.assertEqual(result, {'allow_duplicate_names': True})

        params = XPathMap(parser, items={'allow-duplicate-names': False})
        result = get_serialization_params(params)
        self.assertEqual(result, {'allow_duplicate_names': False})

        params = XPathMap(parser, items={'allow-duplicate-names': 'false'})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'json-node-output-method': 'xml'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'json-node-output-method': 'xml'})

        params = XPathMap(parser, items={'json-node-output-method': True})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        character_map = XPathMap(parser, {'$': '£'})
        params = XPathMap(parser, {'use-character-maps': character_map})
        result = get_serialization_params(params)
        self.assertEqual(result, {'character_map': {'$': '£'}})

        character_map = XPathMap(parser, {'$': 1})
        params = XPathMap(parser, {'use-character-maps': character_map})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        character_map = XPathMap(parser, {'$$': '£'})
        params = XPathMap(parser, {'use-character-maps': character_map})
        with self.assertRaises(ValueError) as ctx:
            get_serialization_params(params)
        self.assertIn('SEPM0016', str(ctx.exception))

        params = XPathMap(parser, items={'standalone': False})
        result = get_serialization_params(params)
        self.assertEqual(result, {'standalone': False})

        params = XPathMap(parser, items={'standalone': True})
        result = get_serialization_params(params)
        self.assertEqual(result, {'standalone': True})

        params = XPathMap(parser, items={'standalone': []})
        result = get_serialization_params(params)
        self.assertEqual(result, {})

        params = XPathMap(parser, items={'standalone': 'no'})
        result = get_serialization_params(params)
        self.assertEqual(result, {'standalone': False})

        params = XPathMap(parser, items={'standalone': 'omit'})
        result = get_serialization_params(params)
        self.assertEqual(result, {})

        params = XPathMap(parser, items={'standalone': ' no '})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        cdata = [
            QName(uri='http://xpath.test/ns', qname='a'),
            QName(uri='http://xpath.test/ns', qname='b'),
            QName(uri='', qname='c')
        ]
        params = XPathMap(parser, items={'cdata-section-elements': cdata})
        result = get_serialization_params(params)
        self.assertEqual(result, {'cdata_section': cdata})

        cdata_array = XPathArray(parser, cdata)
        params = XPathMap(parser, items={'cdata-section-elements': cdata_array})
        result = get_serialization_params(params)
        self.assertEqual(result, {'cdata_section': cdata})

        cdata.append('wrong')
        params = XPathMap(parser, items={'cdata-section-elements': cdata})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

        params = XPathMap(parser, items={'suppress-indentation': QName('', 'foo')})
        result = get_serialization_params(params)
        self.assertEqual(result, {'suppress-indentation': QName('', 'foo')})

        params = XPathMap(parser, items={'suppress-indentation': [QName('', 'foo')]})
        result = get_serialization_params(params)
        self.assertEqual(list(result.values()), [QName('', 'foo')])

        params = XPathMap(parser, items={'suppress-indentation': 'foo'})
        with self.assertRaises(TypeError) as ctx:
            get_serialization_params(params)
        self.assertIn('XPTY0004', str(ctx.exception))

    def test_get_serialization_params_from_element_tree(self):
        namespaces = {'output': "http://www.w3.org/2010/xslt-xquery-serialization"}

        root = ElementTree.XML(dedent("""\
            <output:serialization-parameters
                    xmlns:output="http://www.w3.org/2010/xslt-xquery-serialization">
                <output:method value="xml"/>
                <output:omit-xml-declaration value="yes"/>
                <output:item-separator value="=="/>
            </output:serialization-parameters>"""))

        params = get_node_tree(root, namespaces)
        result = get_serialization_params(params)
        self.assertEqual(result, {'method': 'xml', 'item_separator': '=='})

        root = ElementTree.XML(dedent("""\
            <output:serialization-parameters
                    xmlns:output="http://www.w3.org/2010/xslt-xquery-serialization">
                <output:use-character-maps>
                    <output:character-map character="$" map-string="£"/>
                </output:use-character-maps>
            </output:serialization-parameters>"""))

        params = get_node_tree(root, namespaces)
        result = get_serialization_params(params)
        self.assertEqual(result, {'character_map': {'$': '£'}})

        root = ElementTree.XML(dedent("""\
            <output:serialization-parameters
                    xmlns:output="http://www.w3.org/2010/xslt-xquery-serialization">
                <output:omit-xml-declaration value="no"/>
                <output:standalone value=" no "/>
            </output:serialization-parameters>"""))

        params = get_node_tree(root, namespaces)
        result = get_serialization_params(params)
        self.assertEqual(result, {'standalone': False, 'xml_declaration': True})

    def test_serialize_to_xml_function(self):
        root = ElementTree.XML("<root>1</root>")
        elements = [get_node_tree(root)]
        result = serialize_to_xml(elements)
        self.assertEqual(result, '<root>1</root>')

        root = ElementTree.XML("<root>1</root>")
        elements = [get_node_tree(root)]
        result = serialize_to_xml(elements, xml_declaration=True)
        if sys.version_info < (3, 8):
            self.assertEqual(result, '<root>1</root>')
        else:
            self.assertEqual(result, '<?xml version="1.0" encoding="utf-8"?>\n<root>1</root>')

        cdata = [
            QName(uri='http://xpath.test/ns', qname='a'),
            QName(uri='http://xpath.test/ns', qname='b'),
            QName(uri='', qname='c')
        ]
        result = serialize_to_xml(elements, cdata_section=cdata)
        self.assertEqual(result, '<root>1</root>')

        root1 = ElementTree.XML("<root><c1>£</c1><c2>$</c2></root>")
        root2 = ElementTree.XML("<root><c1>£</c1><c2>€</c2></root>")
        elements = [get_node_tree(root1), get_node_tree(root2)]
        result = serialize_to_xml(elements, character_map={'$': '£'})
        self.assertEqual(result, '<root><c1>£</c1><c2>£</c2></root>'
                                 '<root><c1>£</c1><c2>€</c2></root>')

        root1 = ElementTree.XML("<root1/>")
        root2 = ElementTree.XML("<root2/>")
        elements = [get_node_tree(root1), get_node_tree(root2)]
        result = serialize_to_xml(elements, item_separator='  ')
        self.assertEqual(result, '<root1 />  <root2 />')

        root = ElementTree.XML("<root>1<c/>2<c/>3<c/>4</root>")
        root_node = get_node_tree(root)
        elements = [x for x in root_node.children if isinstance(x, TextNode)]
        result = serialize_to_xml(elements, item_separator='-')
        self.assertEqual(result, '1-2-3-4')

        parser = XPath31Parser()
        elements = [
            XPathArray(parser, [1, 2, 3]),
            XPathArray(parser, [4, 5, 6]),
        ]
        result = serialize_to_xml(elements, item_separator=';')
        self.assertEqual(result, '1;2;3;4;5;6')

        elements = list(range(10))
        result = serialize_to_xml(elements, item_separator=',')
        self.assertEqual(result, '0,1,2,3,4,5,6,7,8,9')

    def test_serialize_to_json_function(self):
        result = serialize_to_json([])
        self.assertEqual(result, 'null')

        result = serialize_to_json(["à"])
        self.assertEqual(result, '"\\u00e0"')

        result = serialize_to_json(["à"], encoding='ascii')
        self.assertEqual(result, '"\\u00e0"')

        root = ElementTree.XML("<root>1</root>")
        elements = [get_node_tree(root)]
        result = serialize_to_json(elements)
        self.assertEqual(result, '"<root>1<\\/root>"')

        document = ElementTree.ElementTree(root)
        elements = [get_node_tree(document)]
        result = serialize_to_json(elements)
        self.assertEqual(result, '"<root>1<\\/root>"')

        root = ElementTree.XML("<root><c1>£</c1><c2>$</c2></root>")
        elements = [get_node_tree(root)]
        result = serialize_to_json(elements)
        self.assertEqual(result, r'"<root><c1>\u00a3<\/c1><c2>$<\/c2><\/root>"')

        root = ElementTree.XML("<root>1<c/>2<c/>3<c/>4</root>")
        elements = [get_node_tree(root)]
        result = serialize_to_json(elements)
        self.assertEqual(result, '"<root>1<c \\/>2<c \\/>3<c \\/>4<\\/root>"')

        root = ElementTree.XML("<root a='bar'>foo</root>")
        root_node = get_node_tree(root)
        result = serialize_to_json(root_node.children)
        self.assertEqual(result, '"foo"')
        result = serialize_to_json(root_node.attributes)
        self.assertEqual(result, '"a=\\"bar\\""')

        comment = ElementTree.Comment(' foo')
        comment_node = CommentNode(content=comment)
        result = serialize_to_json([comment_node])
        self.assertEqual(result, '"<!-- foo-->"')

        pi = ElementTree.ProcessingInstruction('foo bar')
        pi_node = ProcessingInstructionNode(target=pi)
        result = serialize_to_json([pi_node])
        self.assertEqual(result, '"<?foo bar?>"')

        parser = XPath31Parser()

        elements = [XPathArray(parser, [1, 2, 3])]
        result = serialize_to_json(elements)
        self.assertEqual(result, '[1,2,3]')

        elements = [XPathArray(parser, [float('nan'), 2, 3])]
        with self.assertRaises(TypeError) as ctx:
            serialize_to_json(elements)
        self.assertIn('SERE0020', str(ctx.exception))

        elements = [XPathArray(parser, [1, 2, float('inf')])]
        with self.assertRaises(TypeError) as ctx:
            serialize_to_json(elements)
        self.assertIn('SERE0020', str(ctx.exception))

        with self.assertRaises(TypeError) as ctx:
            serialize_to_json([set()])
        self.assertIn('SERE0021', str(ctx.exception))

        elements = list(range(10))
        with self.assertRaises(TypeError) as ctx:
            serialize_to_json(elements)
        self.assertIn('SERE0023', str(ctx.exception))

        elements = [[1, 2, 3]]
        result = serialize_to_json(elements)
        self.assertEqual(result, '[1,2,3]')

        parser = XPath31Parser()
        elements = [XPathMap(parser, [(1, 'one'), (2, 'two'), (3, 'three')])]
        result = serialize_to_json(elements)
        self.assertEqual(result, '{"1":"one","2":"two","3":"three"}')

        elements = [XPathMap(parser, [(1, ['one']), (2, 'two')])]
        result = serialize_to_json(elements)
        self.assertEqual(result, '{"1":["one"],"2":"two"}')

        elements = [XPathMap(parser, [(1, ['one', 'one']), (2, 'two')])]
        result = serialize_to_json(elements)
        self.assertEqual(result, '{"1":["one","one"],"2":"two"}')

        # Previous result was wrong, because ['one', 'one'] is mapped to an array.
        # with self.assertRaises(TypeError) as ctx:
        #    serialize_to_json(elements)
        # self.assertIn('SERE0023', str(ctx.exception))

        elements = [XPathMap(parser, [(QName('', 'one'), 1), ('one', 1)])]
        with self.assertRaises(ValueError) as ctx:
            serialize_to_json(elements)
        self.assertIn('SERE0022', str(ctx.exception))

        result = serialize_to_json(elements, allow_duplicate_names=True)
        self.assertEqual(result, '{"one":1,"one":1}')

        elements = [XPathMap(parser, [(QName('', 'one'), 1), ('two', 2)])]
        result = serialize_to_json(elements)
        self.assertEqual(result, '{"one":1,"two":2}')

        self.assertEqual(serialize_to_json([UntypedAtomic('9.0')]), '"9.0"')
        self.assertEqual(serialize_to_json([decimal.Decimal('9.0')]), '9.0')
        self.assertEqual(serialize_to_json([9.0]), '9.0')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from timeit import timeit
from elementpath.sequences import xlist  # noqa


def run_timeit(stmt='pass', setup='pass', number=1000):
    seconds = timeit(stmt, setup=setup, number=number)
    print("{}: {}s".format(stmt, seconds))


if __name__ == '__main__':
    print('*' * 68)
    print("*** Memory and timing profile of XPath null values alternatives  ***")
    print('*' * 68)
    print()

    NUMBER = 1000
    SETUP = 'from __main__ import xlist'

    print("*** Profile evaluation ***\n")

    run_timeit('[None for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('[[] for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('[xlist() for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('[[] for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('[() for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('xlist([1])', SETUP, NUMBER)
    run_timeit('[xlist() for _ in range(10000)]', SETUP, NUMBER)
    run_timeit('[xlist() for _ in range(10000)]', SETUP, NUMBER)

    print()

    run_timeit('[x for x in range(10000)]', SETUP, NUMBER)
    run_timeit('tuple([x for x in range(10000)])', SETUP, NUMBER)
    run_timeit('list([x for x in range(10000)])', SETUP, NUMBER)
    run_timeit('xlist([x for x in range(10000)])', SETUP, NUMBER)

    run_timeit('\nv = [x for x in range(10000)]\nfor _ in v: pass', SETUP, NUMBER)
    run_timeit('\nv = tuple([x for x in range(10000)])\nfor _ in v: pass', SETUP, NUMBER)
    run_timeit('\nv = list([x for x in range(10000)])\nfor _ in v: pass', SETUP, NUMBER)
    run_timeit('\nv = xlist([x for x in range(10000)])\nfor _ in v: pass', SETUP, NUMBER)



#!/usr/bin/env python
#
# Copyright (c), 2022-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from textwrap import dedent
from xml.etree import ElementTree

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None

try:
    from tests import xpath_test_class
except ImportError:
    import xpath_test_class

from elementpath import XPathContext
# from elementpath.decoder import get_atomic_sequence
from elementpath.datatypes import DateTime


# noinspection PyUnresolvedReferences
@unittest.skipIf(xmlschema is None, "xmlschema library is not installed")
class DecoderTest(xpath_test_class.XPathTestCase):
    etree = ElementTree

    def test_decode_union(self):
        schema = xmlschema.XMLSchema(dedent('''\
            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                <xs:element name="value" type="dateUnion" />
                <xs:simpleType name="dateUnion">
                    <xs:union memberTypes="xs:date xs:dateTime " />
                </xs:simpleType>
            </xs:schema>'''))

        root = self.etree.XML('<value>2007-12-31T12:30:40</value>')
        context = XPathContext(root, schema=schema.xpath_proxy)
        self.check_value('fn:data(.)', [DateTime(2007, 12, 31, 12, 30, 40)], context=context)


@unittest.skipIf(xmlschema is None, "xmlschema library is not installed")
@unittest.skipIf(lxml_etree is None, "lxml library is not installed")
class DecoderLxmlTest(DecoderTest):
    etree = lxml_etree


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from unittest.mock import patch
from textwrap import dedent
import io
import xml.etree.ElementTree as ElementTree

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None
else:
    xmlschema.XMLSchema.meta_schema.build()

from elementpath.etree import is_etree_element, etree_iter_strings, \
    etree_deep_equal, etree_iter_paths
from elementpath.datatypes import UntypedAtomic
from elementpath.xpath_nodes import DocumentNode, ElementNode, AttributeNode, TextNode, \
    NamespaceNode, CommentNode, ProcessingInstructionNode, EtreeElementNode, TextAttributeNode
from elementpath.tree_builders import get_node_tree
from elementpath.xpath_context import XPathContext, XPathSchemaContext


class DummyXsdType:
    name = local_name = None
    xsd_version = '1.0'

    def is_matching(self, name, default_namespace): pass
    def is_empty(self): pass
    def is_simple(self): pass
    def has_simple_content(self): pass
    def has_mixed_content(self): pass
    def is_element_only(self): pass
    def is_key(self): pass
    def is_qname(self): pass
    def is_notation(self): pass
    def decode(self, obj, *args, **kwargs): return int(obj)
    def validate(self, obj, *args, **kwargs): pass


class XPathNodesTest(unittest.TestCase):
    elem = ElementTree.XML('<node a1="10"/>')

    def setUp(self):
        root = ElementTree.Element('root')
        self.context = XPathContext(root)  # Dummy context for creating nodes

    def test_is_etree_element_function(self):
        self.assertTrue(is_etree_element(self.elem))
        self.assertFalse(is_etree_element('text'))
        self.assertFalse(is_etree_element(None))

    def test_elem_iter_strings_function(self):
        root = ElementTree.XML('<A>text1\n<B1>text2</B1>tail1<B2/><B3><C1>text3</C1></B3>tail2</A>')
        result = ['text1\n', 'text2', 'tail1', 'tail2', 'text3']
        self.assertEqual(list(etree_iter_strings(root)), result)

        with patch.multiple(DummyXsdType, has_mixed_content=lambda x: True):
            xsd_type = DummyXsdType()
            typed_root = EtreeElementNode(elem=root)
            setattr(typed_root, 'xsd_type', xsd_type)
            self.assertEqual(list(etree_iter_strings(typed_root.elem)), result)

        norm_result = ['text1', 'text2', 'tail1', 'tail2', 'text3']
        with patch.multiple(DummyXsdType, is_element_only=lambda x: True):
            xsd_type = DummyXsdType()
            typed_root = EtreeElementNode(elem=root)
            setattr(typed_root, 'xsd_type', xsd_type)
            self.assertEqual(list(etree_iter_strings(typed_root.elem, True)), norm_result)

            comment = ElementTree.Comment('foo')
            root[1].append(comment)
            self.assertEqual(list(etree_iter_strings(typed_root.elem, True)), norm_result)

        self.assertEqual(list(etree_iter_strings(root)), result)

    def test_etree_deep_equal_function(self):
        root = ElementTree.XML('<A><B1>10</B1><B2 max="20"/>end</A>')
        self.assertTrue(etree_deep_equal(root, root))

        elem = ElementTree.XML('<A><B1>11</B1><B2 max="20"/>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A><B1>10</B1>30<B2 max="20"/>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A xmlns:ns="tns"><B1>10</B1><B2 max="20"/>end</A>')
        self.assertTrue(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A><B1>10</B1><B2 max="20"><C1/></B2>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

    def test_match_name_method(self):
        attr = AttributeNode('a1', '10', parent=None)
        self.assertTrue(attr.match_name('*'))
        self.assertTrue(attr.match_name('a1'))
        self.assertTrue(attr.match_name('{*}a1'))
        self.assertFalse(attr.match_name('{foo}*'))
        self.assertFalse(attr.match_name('{foo}*'))

        self.assertTrue(
            AttributeNode('{foo}a1', '10').match_name('{foo}*')
        )

        attr = AttributeNode('{http://xpath.test/ns}a1', '10', parent=None)
        self.assertTrue(attr.match_name('{*}a1'))

    def test_node_base_uri(self):
        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="/" />'

        self.assertEqual(EtreeElementNode(ElementTree.XML(xml_test)).base_uri, '/')
        document = ElementTree.parse(io.StringIO(xml_test))

        self.assertIsNone(DocumentNode(document).base_uri, '/')
        self.assertIsNone(EtreeElementNode(self.elem).base_uri)
        self.assertIsNone(TextNode('a text node').base_uri)

        xml_test = dedent("""\
            <?xml version="1.0"?>
            <e1 xml:base="http://example.org/wine/">
              <e2 xml:base="rosé"/>
            </e1>""")

        root_node = get_node_tree(ElementTree.XML(xml_test))
        self.assertEqual(root_node.base_uri, 'http://example.org/wine/')
        self.assertIsInstance(root_node[0], TextNode)
        self.assertEqual(root_node[1].base_uri, 'http://example.org/wine/rosé')

        xml_test = dedent("""\
            <collection xml:base="http://example.test/xpath/ ">
              <item xml:base="urn:isbn:0451450523"/>
              <item xml:base="urn:isan:0000-0000-2CEA-0000-1-0000-0000-Y "/>
              <item xml:base="urn:ISSN:0167-6423 "/>
              <item xml:base=" urn:ietf:rfc:2648 "/>
              <item xml:base="urn:uuid:6e8bc430-9c3a-11d9-9669-0800200c9a66 "/>
            </collection>""")

        root_node = get_node_tree(ElementTree.XML(xml_test))
        self.assertEqual(root_node.base_uri, 'http://example.test/xpath/')
        self.assertIsInstance(root_node[0], TextNode)
        self.assertEqual(root_node[0].base_uri, 'http://example.test/xpath/')
        self.assertEqual(root_node[1].base_uri, 'urn:isbn:0451450523')
        self.assertIsInstance(root_node[2], TextNode)
        self.assertEqual(root_node[3].base_uri, 'urn:isan:0000-0000-2CEA-0000-1-0000-0000-Y')
        self.assertEqual(root_node[5].base_uri, 'urn:ISSN:0167-6423')
        self.assertEqual(root_node[7].base_uri, 'urn:ietf:rfc:2648')
        self.assertEqual(root_node[9].base_uri, 'urn:uuid:6e8bc430-9c3a-11d9-9669-0800200c9a66')

    def test_node_document_uri_function(self):
        node = EtreeElementNode(self.elem)
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="/root" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document)
        self.assertIsNone(node.document_uri)

        node = DocumentNode(document, uri=' http://xpath.test/doc.xml ')
        self.assertEqual(node.document_uri, 'http://xpath.test/doc.xml')

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" ' \
                   'xml:base="http://xpath.test" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document)
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="dir1/dir2" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document, uri="dir1/dir2")
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" ' \
                   'xml:base="http://[xpath.test" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document, uri="http://[xpath.test")
        self.assertIsNone(node.document_uri)

    def test_attribute_nodes(self):
        parent = self.context.root
        attribute = TextAttributeNode('id', '0212349350')

        self.assertEqual(repr(attribute),
                         "TextAttributeNode(name='id', value='0212349350')")
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350'))
        self.assertEqual(attribute.name, 'id')
        self.assertEqual(attribute.value, '0212349350')
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350', parent))

        attribute = AttributeNode('id', '0212349350', parent)
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350', parent))
        self.assertEqual(attribute.value, '0212349350')

        attribute = AttributeNode('value', '10', parent)
        self.assertEqual(repr(attribute), "TextAttributeNode(name='value', value='10')")

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()
            attribute.xsd_type = xsd_type
            self.assertEqual(attribute.value, '10')

    def test_typed_element_nodes(self):
        element = ElementTree.Element('schema')

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()
            context = XPathContext(element)
            context.root.xsd_type = xsd_type
            self.assertTrue(repr(context.root).startswith(
                "EtreeElementNode(elem=<Element 'schema' at 0x"
            ))

    def test_text_nodes(self):
        parent = self.context.root

        # equality if and only is the same instance
        text_node = TextNode('alpha')
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha'))

        text_node = TextNode('alpha', parent)
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha', parent))

        text_node = TextNode('alpha', parent)
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha', parent))

        self.assertEqual(repr(TextNode('alpha')), "TextNode('alpha')")
        text_node = TextNode('alpha', parent)
        self.assertEqual(repr(text_node), "TextNode('alpha')")

        self.assertIsNone(text_node.attributes)
        self.assertIsNone(text_node.children)
        self.assertIsNone(text_node.base_uri)
        self.assertIsNone(text_node.document_uri)
        self.assertIsNone(text_node.is_id)
        self.assertIsNone(text_node.is_idrefs)
        self.assertIsNone(text_node.namespace_nodes)
        self.assertIsNone(text_node.nilled)
        self.assertEqual(text_node.node_kind, 'text')
        self.assertIsNone(text_node.node_name)
        self.assertIsNotNone(text_node.parent)
        self.assertEqual(text_node.string_value, 'alpha')
        self.assertEqual(text_node.typed_value, UntypedAtomic('alpha'))
        self.assertIsNone(text_node.unparsed_entity_public_id('foo'))
        self.assertIsNone(text_node.unparsed_entity_system_id('foo'))

    def test_namespace_nodes(self):
        context = self.context
        namespace = NamespaceNode('tns', 'http://xpath.test/ns')

        self.assertIsNone(namespace.attributes)

        self.assertEqual(repr(namespace),
                         "NamespaceNode(prefix='tns', uri='http://xpath.test/ns')")
        self.assertEqual(namespace.value, 'http://xpath.test/ns')
        self.assertNotEqual(namespace, NamespaceNode('tns', 'http://xpath.test/ns'))
        self.assertEqual(namespace.prefix, 'tns')
        self.assertEqual(namespace.uri, 'http://xpath.test/ns')
        self.assertNotEqual(
            namespace, NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root)
        )

        namespace = NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root)
        self.assertEqual(repr(namespace), "NamespaceNode(prefix='tns', uri='http://xpath.test/ns')")

        self.assertNotEqual(namespace,
                            NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root))
        self.assertEqual(namespace.prefix, 'tns')
        self.assertEqual(namespace.uri, 'http://xpath.test/ns')
        self.assertNotEqual(namespace, NamespaceNode('tns', 'http://xpath.test/ns'))

    def test_node_children_function(self):
        self.assertEqual(EtreeElementNode(self.elem).children, [])
        elem = EtreeElementNode(ElementTree.XML("<A><B1/><B2/></A>"))
        self.assertEqual(elem.children, [x for x in elem])

        document = DocumentNode(ElementTree.parse(io.StringIO("<A><B1/><B2/></A>")))
        self.assertEqual(document.children, [])  # not built document
        EtreeElementNode(document.value.getroot(), document)
        self.assertEqual(document.children, [document.getroot()])

        self.assertIsNone(TextNode('a text node').children)

    def test_node_nilled_property(self):
        xml_test = '<A xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="true" />'
        self.assertTrue(EtreeElementNode(ElementTree.XML(xml_test)).nilled)
        xml_test = '<A xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="false" />'
        self.assertFalse(EtreeElementNode(ElementTree.XML(xml_test)).nilled)
        self.assertFalse(EtreeElementNode(ElementTree.XML('<A />')).nilled)
        self.assertFalse(TextNode('foo').nilled)

    def test_node_kind_property(self):
        document = DocumentNode(ElementTree.parse(io.StringIO(u'<A/>')))
        element = EtreeElementNode(ElementTree.Element('schema'))
        attribute = AttributeNode('id', '0212349350')
        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')
        comment = CommentNode(ElementTree.Comment('nothing important'))
        pi = ProcessingInstructionNode(
            ElementTree.ProcessingInstruction('action', 'nothing to do')
        )
        text = TextNode('betelgeuse')

        self.assertEqual(document.node_kind, 'document')
        self.assertEqual(element.node_kind, 'element')
        self.assertEqual(attribute.node_kind, 'attribute')
        self.assertEqual(namespace.node_kind, 'namespace')
        self.assertEqual(comment.node_kind, 'comment')
        self.assertEqual(pi.node_kind, 'processing-instruction')
        self.assertEqual(text.node_kind, 'text')

    def test_name_property(self):
        root = self.context.root
        attr = AttributeNode('a1', '20')
        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')

        self.assertEqual(root.name, 'root')
        self.assertEqual(attr.name, 'a1')
        self.assertEqual(namespace.name, 'xs')

    def test_path_property(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2 max="10"/></B3></A>')

        context = XPathContext(root)

        self.assertEqual(context.root.path, '/Q{}A[1]')
        self.assertEqual(context.root[0].path, '/Q{}A[1]/Q{}B1[1]')
        self.assertEqual(context.root[0][0].path, '/Q{}A[1]/Q{}B1[1]/Q{}C1[1]')
        self.assertEqual(context.root[1].path, '/Q{}A[1]/Q{}B2[1]')
        self.assertEqual(context.root[2].path, '/Q{}A[1]/Q{}B3[1]')
        self.assertEqual(context.root[2][0].path, '/Q{}A[1]/Q{}B3[1]/Q{}C1[1]')
        self.assertEqual(context.root[2][1].path, '/Q{}A[1]/Q{}B3[1]/Q{}C2[1]')

        attr = context.root[2][1].attributes[0]
        self.assertEqual(attr.path, '/Q{}A[1]/Q{}B3[1]/Q{}C2[1]/@max')

        document = ElementTree.ElementTree(root)
        context = XPathContext(root=document)
        self.assertEqual(context.root[0][2][0].path, '/Q{}A[1]/Q{}B3[1]/Q{}C1[1]')
        self.assertEqual(context.root[0][2][0].extended_path, '/A[1]/B3[1]/C1[1]')

        root = ElementTree.XML('<A><B1>10</B1><B2 min="1"/><B3/></A>')
        context = XPathContext(root)
        with patch.object(DummyXsdType(), 'is_simple', return_value=True) as xsd_type:
            elem = context.root[0]
            elem.xsd_type = xsd_type
            self.assertEqual(elem.path, '/Q{}A[1]/Q{}B1[1]')

        with patch.object(DummyXsdType(), 'is_simple', return_value=True) as xsd_type:
            context = XPathContext(root)
            attr = context.root[1].attributes[0]
            attr.xsd_type = xsd_type
            self.assertEqual(attr.path, '/Q{}A[1]/Q{}B2[1]/@min')

    def test_path_property_with_namespaces(self):
        root = ElementTree.XML('<tns:A xmlns:tns="foo"><B1><C1/></B1><B2/>'
                               '<B3><C1/><C2 max="10"/></B3></tns:A>')

        context = XPathContext(root, namespaces={'tns': 'foo'})
        self.assertEqual(context.root.path, '/Q{foo}A[1]')
        self.assertEqual(context.root.qname_path, '/tns:A[1]')

        self.assertEqual(context.root[0].path, '/Q{foo}A[1]/Q{}B1[1]')
        self.assertEqual(context.root[0][0].qname_path, '/tns:A[1]/B1[1]/C1[1]')

    def test_element_node_iter(self):
        root = ElementTree.XML('<A>text1\n<B1 a="10">text2</B1><B2/><B3><C1>text3</C1></B3></A>')

        context = XPathContext(root)
        expected = [
            context.root, context.root.namespace_nodes[0],
            context.root[0],
            context.root[1], context.root[1].namespace_nodes[0],
            context.root[1].attributes[0], context.root[1][0],
            context.root[2], context.root[2].namespace_nodes[0],
            context.root[3], context.root[3].namespace_nodes[0],
            context.root[3][0], context.root[3][0].namespace_nodes[0],
            context.root[3][0][0]
        ]

        result = list(context.root.iter())
        self.assertEqual(result, expected)

        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        context = XPathContext(root)

        # iter includes also xml namespace nodes
        self.assertEqual(
            list(e.elem for e in context.root.iter() if isinstance(e, ElementNode)),
            list(root.iter())
        )

    def test_document_node_iter(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        doc = ElementTree.ElementTree(root)
        context = XPathContext(doc)

        self.assertEqual(
            list(e.elem for e in context.root.iter() if isinstance(e, ElementNode)),
            list(doc.iter())
        )

    @unittest.skipIf(lxml_etree is None, 'lxml.etree is not installed')
    def test_lazy_attributes_iter__issue_72(self):
        xml = lxml_etree.fromstring('<root id="test"></root>')
        node_tree = get_node_tree(root=xml)

        nodes = list(node for node in node_tree.iter_lazy())
        self.assertEqual(nodes, [node_tree])

        nodes = list(node for node in node_tree.iter())
        self.assertEqual(nodes, [
            node_tree, node_tree.namespace_nodes[0], node_tree.attributes[0]
        ])

        nodes = list(node for node in node_tree.iter_lazy())
        self.assertEqual(nodes, [
            node_tree, node_tree.namespace_nodes[0], node_tree.attributes[0]
        ])

    def test_is_schema_node(self):
        root = ElementTree.XML('<root a="10">text</root>')
        context = XPathContext(root)

        self.assertFalse(context.root.is_schema_node)
        self.assertFalse(context.root.attributes[0].is_schema_node)
        self.assertFalse(context.root.children[0].is_schema_node)

        if xmlschema is not None:
            schema = xmlschema.XMLSchema(dedent("""
                <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                  <xs:element name="elem"/>
                  <xs:attribute name="attr" type="xs:string"/>
                </xs:schema>"""))

            context = XPathSchemaContext(schema)
            self.assertTrue(context.root.is_schema_node)  # Is the schema
            self.assertTrue(context.root.attributes[0].is_schema_node)
            self.assertTrue(context.root.children[0].is_schema_node)

    def test_etree_iter_paths(self):
        root = ElementTree.XML('<a><b1><c1/><c2/></b1><b2/><b3><c3/></b3></a>')
        root[2].append(ElementTree.Comment('a comment'))
        root[2].append(ElementTree.Element('c3'))  # duplicated tag

        items = list(etree_iter_paths(root))
        self.assertEqual(items, [
            (root, '.'), (root[0], './Q{}b1[1]'),
            (root[0][0], './Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], './Q{}b1[1]/Q{}c2[1]'),
            (root[1], './Q{}b2[1]'), (root[2], './Q{}b3[1]'),
            (root[2][0], './Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], './Q{}b3[1]/comment()[1]'),
            (root[2][2], './Q{}b3[1]/Q{}c3[2]')
        ])
        self.assertEqual(list(etree_iter_paths(root, path='')), [
            (root, ''), (root[0], 'Q{}b1[1]'),
            (root[0][0], 'Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], 'Q{}b1[1]/Q{}c2[1]'),
            (root[1], 'Q{}b2[1]'), (root[2], 'Q{}b3[1]'),
            (root[2][0], 'Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], 'Q{}b3[1]/comment()[1]'),
            (root[2][2], 'Q{}b3[1]/Q{}c3[2]')
        ])
        self.assertEqual(list(etree_iter_paths(root, path='/')), [
            (root, '/'), (root[0], '/Q{}b1[1]'),
            (root[0][0], '/Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], '/Q{}b1[1]/Q{}c2[1]'),
            (root[1], '/Q{}b2[1]'), (root[2], '/Q{}b3[1]'),
            (root[2][0], '/Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], '/Q{}b3[1]/comment()[1]'),
            (root[2][2], '/Q{}b3[1]/Q{}c3[2]')
        ])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

def main() -> None:
    from xml.etree.ElementTree import XML
    import elementpath
    from elementpath import get_node_tree
    from elementpath.xpath3 import XPath3Parser

    root = XML('<a><b1/><b2><c1/><c2/></b2><b3/></a>')

    result = elementpath.select(root, '*')
    print(result)

    result = list(elementpath.iter_select(root, '*'))
    print(result)

    selector = elementpath.Selector('*')
    result = selector.select(root)
    print(result)

    result = list(selector.iter_select(root))
    print(result)

    result = elementpath.select(root, 'math:atan(1.0e0)', parser=XPath3Parser)
    print(result)

    root_node = get_node_tree(root)
    result = elementpath.select(root_node, '*')
    print(result)
    assert result == elementpath.select(root, '*')

    try:
        elementpath.select(1, '*')  # type: ignore[arg-type]
    except TypeError:
        pass


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c), 2021, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from timeit import timeit

from elementpath import XPath1Parser
from elementpath.datatypes import Language, UntypedAtomic

def run_timeit(stmt='pass', setup='pass', number=1000):
    seconds = timeit(stmt, setup=setup, number=number)
    print("{}: {}s".format(stmt, seconds))


if __name__ == '__main__':
    print('*' * 62)
    print("*** Memory and timing profile of XPathToken class          ***")
    print("***" + ' ' * 56 + "***")
    print("*** Note: save ~34% of memory with __slots__ (from v2.2.3) ***")
    print('*' * 62)
    print()

    NUMBER = 300000

    print("*** Profile evaluation ***\n")

    run_timeit('UntypedAtomic("foo.bar")', 'from __main__ import UntypedAtomic', NUMBER)

    print()



#!/usr/bin/env python
#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import os.path
import unittest
import pathlib

from elementpath import XPath2Parser, XPathContext
from elementpath.extras.pathnodes import PathElementNode, PathDocumentNode


class PathNodesTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.path = pathlib.Path(__file__)

    def test_build(self):
        node = PathElementNode(self.path)
        self.assertIsInstance(node, PathElementNode)
        self.assertEqual(len(node.tree.elements), len(self.path.parts))
        self.assertIsInstance(node.tree.root_node, PathDocumentNode)

        for _ in node.parent.iter_descendants():
            pass
        self.assertGreater(len(node.tree.elements), 30)

    def test_build_fragment(self):
        node = PathElementNode(self.path, fragment=True)
        self.assertIsInstance(node, PathElementNode)
        self.assertEqual(len(node.tree.elements), 1)
        self.assertIsInstance(node.tree.root_node, PathElementNode)
        self.assertIs(node.tree.root_node, node)
        self.assertIsNone(node.parent)

    def test_attributes(self):
        node = PathElementNode(self.path, fragment=True)
        self.assertGreater(len(node.attributes), 7)

        for attr in node.attributes:
            self.assertEqual(attr.path, f'/{os.path.basename(__file__)}/@{attr.name}')

    def test_find_mtime(self):
        node = PathElementNode(self.path)
        ancestor = node.parent.parent
        assert ancestor is not None
        path = './/element(dist)/*'
        token = XPath2Parser(strict=False).parse(path)
        context = XPathContext(ancestor)

        for path in token.select(context):
            pass  # print(path)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2022-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from xml.etree import ElementTree

from elementpath import XPath2Parser
from elementpath.xpath_nodes import EtreeElementNode
from elementpath.tree_builders import get_node_tree
from elementpath.compare import deep_equal, deep_compare, get_key_function


class CompareTest(unittest.TestCase):

    def test_deep_equal_function(self):
        parser = XPath2Parser()
        token = parser.parse('true()')

        with self.assertRaises(TypeError):
            deep_equal([token], [1])

        with self.assertRaises(TypeError):
            deep_equal([1], [token])

        self.assertTrue(deep_equal([1], [1]))
        self.assertFalse(deep_equal([1], [2]))
        self.assertFalse(deep_equal([1, 1], [1]))
        self.assertFalse(deep_equal([1], [1, 1]))

        root = ElementTree.Element('root')
        elem = ElementTree.Element('elem')
        element = EtreeElementNode(elem)
        self.assertTrue(deep_equal([element], [element]))
        self.assertFalse(deep_equal([1], [element]))
        self.assertFalse(deep_equal([EtreeElementNode(root)], [element]))

        root = ElementTree.XML('<root a="1">text<child b="2"/>tail<child/></root>')
        element = get_node_tree(root)
        document = get_node_tree(ElementTree.ElementTree(root))
        self.assertTrue(deep_equal([element], [element]))
        self.assertTrue(deep_equal([document], [document]))

        root1 = ElementTree.XML('<root a="1">text<child b="2"/>tail</root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertFalse(deep_equal([element], [element1]))
        self.assertFalse(deep_equal([document], [document1]))

        root1 = ElementTree.XML('<root a="1"><child b="2"/>tail<child/></root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertFalse(deep_equal([element], [element1]))
        self.assertFalse(deep_equal([document], [document1]))

        root1 = ElementTree.XML('<root a="1">text<child b="3"/>tail<child/></root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertFalse(deep_equal([element], [element1]))
        self.assertFalse(deep_equal([document], [document1]))

    def test_deep_compare(self):
        parser = XPath2Parser()
        token = parser.parse('true()')

        with self.assertRaises(TypeError):
            deep_compare([token], [1])

        with self.assertRaises(TypeError):
            deep_compare([1], [token])

        self.assertEqual(deep_compare([1], [1]), 0)
        self.assertEqual(deep_compare([1], [2]), -1)
        self.assertEqual(deep_compare([1, 1], [1]), 1)
        self.assertEqual(deep_compare([1], [1, 1]), -1)

        root = ElementTree.Element('root')
        elem = ElementTree.Element('elem')
        element = EtreeElementNode(elem)
        self.assertEqual(deep_compare([element], [element]), 0)

        with self.assertRaises(TypeError):
            deep_compare([1], [element])

        self.assertEqual(deep_compare([EtreeElementNode(root)], [element]), 1)

        root = ElementTree.XML('<root a="1">text<child b="2"/>tail<child/></root>')
        element = get_node_tree(root)
        document = get_node_tree(ElementTree.ElementTree(root))
        self.assertEqual(deep_compare([element], [element]), 0)
        self.assertEqual(deep_compare([document], [document]), 0)

        root1 = ElementTree.XML('<root a="1">text<child b="2"/>tail</root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertEqual(deep_compare([element], [element1]), -1)
        self.assertEqual(deep_compare([document], [document1]), -1)

        root1 = ElementTree.XML('<root a="1"><child b="2"/>tail<child/></root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertEqual(deep_compare([element], [element1]), 1)
        self.assertEqual(deep_compare([document], [document1]), 1)

        root1 = ElementTree.XML('<root a="1">text<child b="3"/>tail<child/></root>')
        element1 = get_node_tree(root1)
        document1 = get_node_tree(ElementTree.ElementTree(root1))
        self.assertEqual(deep_compare([element], [element1]), -1)
        self.assertEqual(deep_compare([document], [document1]), -1)

    def test_key_function(self):
        key_function = get_key_function()
        result = sorted([2, 1], key=key_function)
        self.assertEqual(result, [1, 2])

        result = sorted([2, 1, 0], key=key_function)
        self.assertEqual(result, [0, 1, 2])

        result = sorted([2, 10, 7], key=key_function)
        self.assertEqual(result, [2, 7, 10])

        with self.assertRaises(TypeError) as cm:
            sorted(['2', 1, 0], key=key_function)
        self.assertIn('XPTY0004', str(cm.exception))

        result = sorted(['2', '10', '7'], key=key_function)
        self.assertEqual(result, ['10', '2', '7'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
# flake8: noqa

def profile_memory(func):
    def wrapper(*a, **kw):
        mem = this_process.memory_info().rss
        result = func(*a, **kw)
        mem = this_process.memory_info().rss - mem
        print("Memory usage by %s(): %.2f MB (%d)" % (func.__name__, mem / 1024 ** 2, mem))
        return result

    return wrapper


# noinspection PyUnresolvedReferences
@profile_memory
def elementpath_deps_memory_usage():
    # Memory relevant standard library imports
    import pathlib
    import decimal
    import calendar
    import xml.etree.ElementTree
    import unicodedata


@profile_memory
def elementpath_memory_usage():
    # elementpath imports
    #
    # Note: comments out all subpackages imports in elementpath/__init__.py
    # to put in evidence the memory consumption of each subpackage.
    #
    import elementpath

@profile_memory
def elementpath_subpackages_memory_usage():
    import elementpath.regex
    import elementpath.datatypes
    import elementpath.xpath_nodes
    import elementpath.xpath_context
    import elementpath.xpath_tokens
    import elementpath.xpath1
    import elementpath.xpath2


@profile_memory
def elementpath_optional_memory_usage():
    # Optional elementpath imports
    import elementpath.xpath30
    import elementpath.xpath31


if __name__ == '__main__':
    import os
    import psutil

    this_process = psutil.Process(os.getpid())

    elementpath_deps_memory_usage()
    elementpath_memory_usage()
    elementpath_subpackages_memory_usage()
    elementpath_optional_memory_usage()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
import io
import xml.etree.ElementTree as ElementTree
from textwrap import dedent

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None
else:
    xmlschema.XMLSchema.meta_schema.build()

from elementpath.tree_builders import build_node_tree, \
    build_lxml_node_tree, build_schema_node_tree
from elementpath.xpath_nodes import ElementNode, \
    DocumentNode, TextNode, CommentNode, ProcessingInstructionNode


XML_DATA = """\
<?xml version='1.0' encoding='UTF8'?>
<?xml-model type="application/xml"?>
<!-- Document comment -->
<root xmlns:tns="http://elementpath.test/ns">
    <!-- Root comment -->
    <child1>
        child1 text1
        <!-- Child 1 comment -->
        <elem1 xmlns:xml="http://www.w3.org/XML/1998/namespace" a1="value1"/>
        <elem2 a2="value2"/>
        child1 text2
    </child1>
    <child2/>
    <tns:other/>
    <?PI-target PI content?>
</root>
<?xml-model type="application/xml"?>
<!-- Document comment 2 -->
"""


class TreeBuildersTest(unittest.TestCase):
    namespaces = {'tns': "http://elementpath.test/ns"}

    def test_build_node_tree_with_element(self):
        root = ElementTree.XML(XML_DATA)
        node = build_node_tree(root, self.namespaces)

        self.assertIsInstance(node, ElementNode)
        self.assertIn(root, node.elements)
        self.assertEqual(len(node.children), 7)
        self.assertIsInstance(node.children[0], TextNode)
        self.assertIsInstance(node.children[1], ElementNode)
        self.assertIsInstance(node.children[2], TextNode)
        self.assertIsInstance(node.children[3], ElementNode)
        self.assertIsInstance(node.children[4], TextNode)
        self.assertIsInstance(node.children[5], ElementNode)
        self.assertIsInstance(node.children[6], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

    def test_build_node_tree_with_element_tree(self):
        root = ElementTree.parse(io.StringIO(XML_DATA))
        node = build_node_tree(root, self.namespaces)

        self.assertIsInstance(node, DocumentNode)
        self.assertEqual(node.position, 1)
        self.assertEqual(len(node.children), 1)

        self.assertIsInstance(node[0], ElementNode)
        self.assertEqual(node[0].position, 2)
        self.assertEqual(len(node[0].children), 7)
        self.assertIsInstance(node[0].children[0], TextNode)
        self.assertIsInstance(node[0].children[1], ElementNode)
        self.assertIsInstance(node[0].children[2], TextNode)
        self.assertIsInstance(node[0].children[3], ElementNode)
        self.assertIsInstance(node[0].children[4], TextNode)
        self.assertIsInstance(node[0].children[5], ElementNode)
        self.assertIsInstance(node[0].children[6], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

    def test_build_node_tree_with_comments_and_pis(self):
        parser = ElementTree.XMLParser(
            target=ElementTree.TreeBuilder(
                insert_comments=True, insert_pis=True
            )
        )
        root = ElementTree.XML(XML_DATA, parser=parser)
        node = build_node_tree(root, self.namespaces)
        self.assertIsInstance(node, ElementNode)

        self.assertEqual(len(node.children), 11)
        self.assertIsInstance(node.children[0], TextNode)
        self.assertIsInstance(node.children[1], CommentNode)
        self.assertIsInstance(node.children[2], TextNode)
        self.assertIsInstance(node.children[3], ElementNode)
        self.assertIsInstance(node.children[4], TextNode)
        self.assertIsInstance(node.children[5], ElementNode)
        self.assertIsInstance(node.children[6], TextNode)
        self.assertIsInstance(node.children[7], ElementNode)
        self.assertIsInstance(node.children[8], TextNode)
        self.assertIsInstance(node.children[9], ProcessingInstructionNode)
        self.assertIsInstance(node.children[10], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

    @unittest.skipIf(lxml_etree is None, "lxml library is not installed")
    def test_build_lxml_node_tree_with_element(self):
        root = lxml_etree.XML(XML_DATA.encode('utf-8'))
        node = build_lxml_node_tree(root)

        self.assertIsInstance(node, DocumentNode)
        self.assertEqual(node.position, 1)
        self.assertIsNotNone(node.document)
        self.assertEqual(len(node.children), 5)
        self.assertIsInstance(node.children[0], ProcessingInstructionNode)
        self.assertIsInstance(node.children[1], CommentNode)
        self.assertIsInstance(node.children[2], ElementNode)
        self.assertIsInstance(node.children[3], ProcessingInstructionNode)
        self.assertIsInstance(node.children[4], CommentNode)

        self.assertIsInstance(node[2], ElementNode)
        self.assertEqual(node[2].position, 4)
        self.assertEqual(len(node[2].children), 11)
        self.assertIsInstance(node[2].children[0], TextNode)
        self.assertIsInstance(node[2].children[1], CommentNode)
        self.assertIsInstance(node[2].children[2], TextNode)
        self.assertIsInstance(node[2].children[3], ElementNode)
        self.assertIsInstance(node[2].children[4], TextNode)
        self.assertIsInstance(node[2].children[5], ElementNode)
        self.assertIsInstance(node[2].children[6], TextNode)
        self.assertIsInstance(node[2].children[5], ElementNode)
        self.assertIsInstance(node[2].children[6], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

        node = build_lxml_node_tree(root[1])
        self.assertIsInstance(node, ElementNode)
        self.assertEqual(len(node.children), 7)
        self.assertIsInstance(node.children[0], TextNode)
        self.assertIsInstance(node.children[1], CommentNode)
        self.assertIsInstance(node.children[2], TextNode)
        self.assertIsInstance(node.children[3], ElementNode)
        self.assertIsInstance(node.children[4], TextNode)
        self.assertIsInstance(node.children[5], ElementNode)
        self.assertIsInstance(node.children[6], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

    @unittest.skipIf(lxml_etree is None, "lxml library is not installed")
    def test_build_lxml_node_tree_with_element_tree(self):
        root = lxml_etree.parse(io.BytesIO(XML_DATA.encode('utf-8')))
        node = build_lxml_node_tree(root)

        self.assertIsInstance(node, DocumentNode)
        self.assertEqual(node.position, 1)
        self.assertIs(node.document, root)
        self.assertEqual(len(node.children), 5)
        self.assertIsInstance(node.children[0], ProcessingInstructionNode)
        self.assertIsInstance(node.children[1], CommentNode)
        self.assertIsInstance(node.children[2], ElementNode)
        self.assertIsInstance(node.children[3], ProcessingInstructionNode)
        self.assertIsInstance(node.children[4], CommentNode)

        self.assertIsInstance(node[2], ElementNode)
        self.assertEqual(node[2].position, 4)
        self.assertEqual(len(node[2].children), 11)
        self.assertIsInstance(node[2].children[0], TextNode)
        self.assertIsInstance(node[2].children[1], CommentNode)
        self.assertIsInstance(node[2].children[2], TextNode)
        self.assertIsInstance(node[2].children[3], ElementNode)
        self.assertIsInstance(node[2].children[4], TextNode)
        self.assertIsInstance(node[2].children[5], ElementNode)
        self.assertIsInstance(node[2].children[6], TextNode)
        self.assertIsInstance(node[2].children[5], ElementNode)
        self.assertIsInstance(node[2].children[6], TextNode)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

        root = lxml_etree.ElementTree()
        self.assertIsNone(root.getroot())
        node = build_lxml_node_tree(root)

        self.assertIsInstance(node, DocumentNode)
        self.assertEqual(node.position, 1)
        self.assertIs(node.document, root)
        self.assertEqual(len(node.children), 0)

        for k, node in enumerate(node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

    @unittest.skipIf(xmlschema is None, "xmlschema library is not installed!")
    def test_build_schema_node_tree(self):
        schema = xmlschema.XMLSchema(dedent("""\n
            <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                <xs:element name="root">
                    <xs:complexType>
                        <xs:sequence>
                            <xs:element name="elem1"/>
                            <xs:element name="elem2"/>
                            <xs:element name="elem3"/>
                            <xs:element ref="root"/>
                        </xs:sequence>
                    </xs:complexType>
                </xs:element>
            </xs:schema>"""))

        root_node = build_schema_node_tree(schema)

        self.assertIs(root_node.elem, schema)
        self.assertIsInstance(root_node.elements, dict)
        self.assertIn(schema, root_node.elements)

        for node in root_node.elements.values():
            self.assertIs(node.elements, root_node.elements)
        self.assertEqual(len(root_node.elements), 7)

        global_elements = []
        root_node = build_schema_node_tree(schema, global_elements=global_elements)
        self.assertIs(root_node.elem, schema)
        self.assertIn(root_node, global_elements)

        for node in root_node.elements.values():
            self.assertIs(node.elements, root_node.elements)
        self.assertEqual(len(root_node.elements), 7)

        root_node = build_schema_node_tree(schema.elements['root'])
        self.assertIs(root_node.elem, schema.elements['root'])

        self.assertIsInstance(root_node.elements, dict)
        for node in root_node.elements.values():
            self.assertIs(node.elements, root_node.elements)
        self.assertEqual(len(root_node.elements), 6)

    def test_document_order__issue_079(self):
        xml_source = '<A>10<B a="2" b="3">11</B>12<B a="2"/>13<B>14</B></A>'
        root = ElementTree.XML(xml_source)
        root_node = build_node_tree(root)

        for k, node in enumerate(root_node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

        if lxml_etree is not None:
            root = lxml_etree.XML(xml_source)
            root_node = build_lxml_node_tree(root)

            for k, node in enumerate(root_node.iter(), start=1):
                self.assertEqual(k, node.position, msg=node)

        xml_source = '<A>10<B a="2" b="3"><C>11</C></B>12<B a="2"/>13<B>14</B></A>'
        root = ElementTree.XML(xml_source)
        root_node = build_node_tree(root)

        for k, node in enumerate(root_node.iter(), start=1):
            self.assertEqual(k, node.position, msg=node)

        if lxml_etree is not None:
            root = lxml_etree.XML(xml_source)
            root_node = build_lxml_node_tree(root)

            for k, node in enumerate(root_node.iter(), start=1):
                self.assertEqual(k, node.position, msg=node)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from copy import copy
from unittest.mock import patch
import xml.etree.ElementTree as ElementTree

try:
    import lxml.etree as lxml_etree
    import lxml.html as lxml_html
except ImportError:
    lxml_etree = None
    lxml_html = None

from elementpath import XPathContext, DocumentNode, ElementNode, datatypes, \
    select, get_node_tree, TextNode


class DummyXsdType:
    name = local_name = None
    xsd_version = '1.0'

    @property
    def root_type(self): return self
    @property
    def simple_type(self): return self
    def is_matching(self, name, default_namespace): pass
    def is_empty(self): pass
    def is_simple(self): pass
    def has_simple_content(self): pass
    def has_mixed_content(self): pass
    def is_element_only(self): pass
    def is_key(self): pass
    def is_qname(self): pass
    def is_notation(self): pass
    def decode(self, obj, *args, **kwargs): pass
    def validate(self, obj, *args, **kwargs): pass


class XPathContextTest(unittest.TestCase):
    root = ElementTree.XML('<author>Dickens</author>')

    def test_invalid_initialization(self):
        self.assertRaises(TypeError, XPathContext, None)

        with self.assertRaises(TypeError):
            XPathContext(item=(1,))

    def test_timezone_argument(self):
        context = XPathContext(self.root)
        self.assertIsNone(context.timezone)
        context = XPathContext(self.root, timezone='Z')
        self.assertIsInstance(context.timezone, datatypes.Timezone)

    def test_repr(self):
        self.assertEqual(repr(XPathContext(self.root)), f"XPathContext(root={self.root})")
        self.assertEqual(repr(XPathContext(item=self.root)), f"XPathContext(item={self.root})")
        self.assertEqual(repr(XPathContext(item=9.0)), "XPathContext(item=9.0)")

    def test_copy(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        context = XPathContext(root)
        self.assertIsInstance(copy(context), XPathContext)
        self.assertIsNot(copy(context), context)

    @unittest.skipIf(lxml_etree is None, 'lxml library is not installed')
    def test_etree_property(self):
        root = ElementTree.XML('<root/>')
        context = XPathContext(root)
        self.assertEqual(context.etree.__name__, 'xml.etree.ElementTree')
        self.assertEqual(context.etree.__name__, 'xml.etree.ElementTree')  # property caching

        root = lxml_etree.XML('<root/>')
        context = XPathContext(root)
        self.assertEqual(context.etree.__name__, 'lxml.etree')
        self.assertEqual(context.etree.__name__, 'lxml.etree')

    def test_context_root_type(self):
        root = ElementTree.XML('<root/>')
        context = XPathContext(root)
        self.assertTrue(context.is_document())
        self.assertIsInstance(context.root, ElementNode)
        self.assertIsInstance(context.document, DocumentNode)
        self.assertFalse(context.is_fragment())
        self.assertFalse(context.is_rooted_subtree())

        root = ElementTree.XML('<root/>')
        context = XPathContext(root, fragment=True)
        self.assertFalse(context.is_document())
        self.assertIsInstance(context.root, ElementNode)
        self.assertIsNone(context.document)
        self.assertIsNone(context.root.parent)
        self.assertTrue(context.is_fragment())
        self.assertFalse(context.is_rooted_subtree())

        root = ElementTree.XML('<root><child/></root>')
        context = XPathContext(root[0], fragment=True)

        self.assertFalse(context.is_document())
        self.assertIsInstance(context.root, ElementNode)
        self.assertIsNone(context.root.parent)
        self.assertIsNone(context.document)
        self.assertTrue(context.is_fragment())
        self.assertFalse(context.is_rooted_subtree())

    def test_no_root(self):
        with self.assertRaises(TypeError) as ctx:
            XPathContext()
        self.assertEqual(str(ctx.exception),
                         "Missing both the root node and the context item!")

        context = XPathContext(item=7)
        self.assertIsNone(context.root)
        self.assertEqual(context.item, 7)

        self.assertEqual(list(context.iter_self()), [7])
        self.assertEqual(list(context.iter_children_or_self()), [])
        self.assertEqual(list(context.iter_attributes()), [])
        self.assertEqual(list(context.iter_descendants()), [])
        self.assertEqual(list(context.iter_parent()), [])
        self.assertEqual(list(context.iter_preceding()), [])
        self.assertEqual(list(context.iter_followings()), [])
        self.assertEqual(list(context.iter_ancestors()), [])
        self.assertEqual(context.item, 7)

        root = ElementTree.XML('<root><child1/><child2/></root>')
        root_node = get_node_tree(root)
        context = XPathContext(item=root_node)
        self.assertEqual(context.item, root_node)

        self.assertEqual(list(context.iter_self()), [root_node])
        self.assertEqual(list(context.iter_children_or_self()), root_node[:])
        self.assertEqual(list(context.iter_attributes()), [])
        self.assertEqual(list(context.iter_descendants()),
                         [root_node, root_node[0], root_node[1]])
        self.assertEqual(list(context.iter_parent()), [])
        self.assertEqual(list(context.iter_preceding()), [])
        self.assertEqual(list(context.iter_followings()), [])
        self.assertEqual(list(context.iter_ancestors()), [])
        self.assertEqual(context.item, root_node)

        context = XPathContext(item=root_node[0])
        self.assertEqual(context.item, root_node[0])

        self.assertEqual(list(context.iter_self()), [root_node[0]])
        self.assertEqual(list(context.iter_children_or_self()), [])
        self.assertEqual(list(context.iter_attributes()), [])
        self.assertEqual(list(context.iter_descendants()), [root_node[0]])
        self.assertEqual(list(context.iter_parent()), [root_node])
        self.assertEqual(list(context.iter_preceding()), [])
        self.assertEqual(list(context.iter_followings()), [root_node[1]])
        self.assertEqual(list(context.iter_ancestors()), [root_node])
        self.assertEqual(context.item, root_node[0])

        context = XPathContext(item=root_node[1])
        self.assertEqual(context.item, root_node[1])

        self.assertEqual(list(context.iter_self()), [root_node[1]])
        self.assertEqual(list(context.iter_children_or_self()), [])
        self.assertEqual(list(context.iter_attributes()), [])
        self.assertEqual(list(context.iter_descendants()), [root_node[1]])
        self.assertEqual(list(context.iter_parent()), [root_node])
        self.assertEqual(list(context.iter_preceding()), [root_node[0]])
        self.assertEqual(list(context.iter_followings()), [])
        self.assertEqual(list(context.iter_ancestors()), [root_node])
        self.assertEqual(context.item, root_node[1])

    def test_default_collection(self):
        node = TextNode('hello world!')

        context = XPathContext(self.root, default_collection=1)
        self.assertEqual(context.default_collection, [])
        context = XPathContext(self.root, default_collection=[node])
        self.assertEqual(context.default_collection, [node])

    def test_is_principal_node_kind(self):
        root = ElementTree.XML('<A a1="10" a2="20"/>')
        context = XPathContext(root)
        self.assertTrue(hasattr(context.item.value, 'tag'))
        self.assertTrue(context.is_principal_node_kind())
        context.item = context.root.attributes[0]
        self.assertFalse(context.is_principal_node_kind())
        context.axis = 'attribute'
        self.assertTrue(context.is_principal_node_kind())

    def test_iter_product(self):
        context = XPathContext(self.root)

        def sel1(_context):
            yield from range(2)

        def sel2(_context):
            yield from range(3)

        expected = [(0, 0), (0, 1), (0, 2), (1, 0), (1, 1), (1, 2)]

        self.assertEqual(list(context.iter_product([sel1, sel2])), expected)
        self.assertEqual(context.variables, {})

        self.assertEqual(list(context.iter_product([sel1, sel2], [])), expected)
        self.assertEqual(context.variables, {})

        self.assertEqual(list(context.iter_product([sel1, sel2], ['a', 'b'])), expected)
        self.assertEqual(context.variables, {'a': 1, 'b': 2})

        context.variables = {'a': 0, 'b': 0}
        self.assertEqual(list(context.iter_product([sel1, sel2], ['a', 'b'])), expected)
        self.assertEqual(context.variables, {'a': 1, 'b': 2})

        context.variables = {'a': 0, 'b': 0}
        self.assertEqual(list(context.iter_product([sel1, sel2], ['a'])), expected)
        self.assertEqual(context.variables, {'a': 1, 'b': 0})

        context.variables = {'a': 0, 'b': 0}
        self.assertEqual(list(context.iter_product([sel1, sel2], ['c', 'b'])), expected)
        self.assertEqual(context.variables, {'a': 0, 'b': 2, 'c': 1})

        context.variables = {'a': 0, 'b': 0}
        self.assertEqual(list(context.iter_product([sel1, sel2], ['b'])), expected)
        self.assertEqual(context.variables, {'a': 0, 'b': 1})

    def test_iter_attributes(self):
        root = ElementTree.XML('<A a1="10" a2="20"/>')
        context = XPathContext(root)
        attributes = context.root.attributes

        self.assertEqual(len(attributes), 2)
        self.assertEqual(list(context.iter_attributes()), attributes)

        context.item = attributes[0]
        self.assertEqual(list(context.iter_attributes()), attributes[:1])

        with patch.object(DummyXsdType(), 'has_simple_content', return_value=True) as xsd_type:
            context = XPathContext(root)
            context.root.xsd_type = xsd_type
            self.assertEqual(list(context.iter_attributes()), context.root.attributes)
            self.assertNotEqual(attributes, context.root.attributes)

    def test_iter_children_or_self(self):
        doc = ElementTree.ElementTree(self.root)
        context = XPathContext(doc)
        self.assertIsInstance(context.root, DocumentNode)
        self.assertIsInstance(context.root[0], ElementNode)

        self.assertEqual(list(e.value for e in context.iter_children_or_self()), [self.root])

        context.item = context.root[0]  # root element
        self.assertEqual(list(context.iter_children_or_self()),
                         [context.root[0].children[0]])

        context.item = context.root  # document node
        self.assertEqual(list(e.value for e in context.iter_children_or_self()), [self.root])

    def test_iter_parent(self):
        root = ElementTree.XML('<A a1="10" a2="20"/>')
        context = XPathContext(root, item=None)
        self.assertEqual(list(context.iter_parent()), [])

        context = XPathContext(root)
        self.assertEqual(list(context.iter_parent()), [])

        with patch.object(DummyXsdType(), 'has_simple_content', return_value=True) as xsd_type:
            context = XPathContext(root, item=root)
            context.root.xsd_type = xsd_type
            self.assertEqual(list(context.iter_parent()), [])

        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        context = XPathContext(root, item=None)
        self.assertEqual(list(context.iter_parent()), [])

        context = XPathContext(root, item=root[2][0])
        self.assertEqual(list(e.obj for e in context.iter_parent()), [root[2]])

        with patch.object(DummyXsdType(), 'is_empty', return_value=True) as xsd_type:
            context = XPathContext(root, item=root[2][0])
            context.root[2][0].xsd_type = xsd_type
            self.assertEqual(list(e.obj for e in context.iter_parent()), [root[2]])

    def test_iter_siblings(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/></B3><B4/><B5/></A>')

        context = XPathContext(root)
        self.assertEqual(list(context.iter_siblings()), [])

        context = XPathContext(root, item=root[2])
        self.assertEqual(list(e.obj for e in context.iter_siblings()), list(root[3:]))

        with patch.object(DummyXsdType(), 'is_element_only', return_value=True) as xsd_type:
            context = XPathContext(root, item=root[2])
            context.root[2].xsd_type = xsd_type
            self.assertEqual(list(e.obj for e in context.iter_siblings()), list(root[3:]))

        context = XPathContext(root, item=root[2])
        self.assertEqual(
            list(e.obj for e in context.iter_siblings('preceding-sibling')), list(root[:2])
        )

        with patch.object(DummyXsdType(), 'is_element_only', return_value=True) as xsd_type:
            context = XPathContext(root, item=root[2])
            context.root[2].xsd_type = xsd_type
            self.assertEqual(
                list(e.obj for e in context.iter_siblings('preceding-sibling')), list(root[:2])
            )

    @unittest.skipIf(lxml_etree is None, 'lxml library is not installed')
    def test_iter_siblings__issue_44(self):
        root = lxml_etree.XML('<root>text 1<!-- comment -->text 2<!-- comment --> text 3</root>')
        result = select(root, 'node()[1]/following-sibling::node()')
        self.assertEqual(result, [root[0], 'text 2', root[1], ' text 3'])
        self.assertEqual(result, root.xpath('node()[1]/following-sibling::node()'))

    @unittest.skipIf(lxml_etree is None, 'lxml library is not installed')
    def test_set_context_root__issue_71(self):
        root = lxml_etree.XML('<root><child1/><child2/></root>')
        self.assertIsNone(root.getparent())

        context = XPathContext(root)
        self.assertIs(context.root.obj, root)
        self.assertIsInstance(context.document, DocumentNode)

        parser = lxml_html.HTMLParser()
        root = lxml_html.fromstring('<root><child1/><child2/></root>', parser=parser)
        self.assertIsNotNone(root.getparent())

        context = XPathContext(root)
        self.assertIs(context.root.obj, root)
        self.assertIsInstance(context.document, DocumentNode)

    def test_iter_descendants(self):
        root = ElementTree.XML('<A a1="10" a2="20"><B1/><B2/></A>')
        context = XPathContext(root)
        attr = context.root.attributes[0]

        self.assertEqual(list(e.value for e in context.iter_descendants()),
                         [root, root[0], root[1]])

        context.item = attr
        self.assertEqual(list(context.iter_descendants(axis='descendant')), [])

        context.item = attr
        self.assertEqual(list(context.iter_descendants()), [attr])

        with patch.object(DummyXsdType(), 'has_mixed_content', return_value=True) as xsd_type:
            context = XPathContext(root, item=root)
            context.root.xsd_type = xsd_type
            self.assertEqual(
                list(e.value for e in context.iter_descendants()), [root, root[0], root[1]]
            )

    def test_iter_ancestors(self):
        root = ElementTree.XML('<A a1="10" a2="20"><B1/><B2/></A>')
        context = XPathContext(root)
        attr = context.root.attributes[0]

        self.assertEqual(list(context.iter_ancestors()), [])

        context.item = attr
        self.assertEqual(list(context.iter_ancestors()), [context.root])

        result = list(e.value for e in XPathContext(root, item=root[1]).iter_ancestors())
        self.assertEqual(result, [root])
        with patch.object(DummyXsdType(), 'has_mixed_content', return_value=True) as xsd_type:
            context = XPathContext(root, item=root[1])
            context.root[1].xsd_type = xsd_type
            self.assertEqual(list(context.iter_ancestors()), [context.root])

    def test_iter_preceding(self):
        root = ElementTree.XML('<A a1="10" a2="20"/>')
        context = XPathContext(root, item=None)
        self.assertEqual(list(context.iter_preceding()), [])

        context = XPathContext(root)
        self.assertEqual(list(context.iter_preceding()), [])

        with patch.object(DummyXsdType(), 'has_simple_content', return_value=True) as xsd_type:
            context = XPathContext(root, item=root)
            context.root.xsd_type = xsd_type
            self.assertEqual(list(context.iter_preceding()), [])

        context = XPathContext(root, item='text')
        self.assertEqual(list(context.iter_preceding()), [])

        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        context = XPathContext(root, item=root[2][1])
        self.assertEqual(list(e.obj for e in context.iter_preceding()),
                         [root[0], root[0][0], root[1], root[2][0]])

    def test_iter_following(self):
        root = ElementTree.XML('<A a="1"><B1><C1/></B1><B2/><B3><C1/></B3><B4/><B5/></A>')

        context = XPathContext(root)
        self.assertEqual(list(context.iter_followings()), [])

        context = XPathContext(root)
        context.item = context.root.attributes[0]
        self.assertEqual(list(context.iter_followings()), [])

        context = XPathContext(root, item=root[2])
        self.assertEqual(list(e.obj for e in context.iter_followings()), list(root[3:]))

        context = XPathContext(root, item=root[1])
        result = [root[2], root[2][0], root[3], root[4]]
        self.assertEqual(list(e.obj for e in context.iter_followings()), result)

        with patch.object(DummyXsdType(), 'has_mixed_content', return_value=True) as xsd_type:
            context = XPathContext(root, item=root[1])
            context.root[1].xsd_type = xsd_type
            self.assertEqual(list(e.obj for e in context.iter_followings()), result)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest

from elementpath.sequences import xlist, empty_sequence, \
    sequence_concat, sequence_count, iterate_sequence, XSequence


class ListSequencesTest(unittest.TestCase):

    def test_string_repr(self):
        self.assertEqual(repr(xlist([])), '[]')
        self.assertEqual(repr(xlist([1, 2, 3])), '[1, 2, 3]')
        self.assertEqual(repr(xlist((x for x in [1, 2, 3]))), '[1, 2, 3]')
        self.assertEqual(repr(xlist([1])), '[1]')

        self.assertEqual(str(xlist([1, 2, 3])), '[1, 2, 3]')
        self.assertEqual(str(xlist([1])), '[1]')
        self.assertEqual(str(xlist()), '[]')

        self.assertEqual(repr(xlist([1, 2, 3])), '[1, 2, 3]')
        self.assertEqual(repr(xlist([1])), '[1]')
        self.assertEqual(repr(xlist()), '[]')

    def test_empty_sequence(self):
        self.assertNotEqual(empty_sequence(), ())
        self.assertEqual(empty_sequence(), [])
        self.assertEqual(empty_sequence(), xlist())
        self.assertNotIsInstance(empty_sequence(), xlist)

    def test_sequence_concat(self):
        self.assertListEqual(sequence_concat([], []), [])
        self.assertListEqual(sequence_concat([1], []), [1])
        self.assertTrue(sequence_concat([1], []) == 1)
        self.assertListEqual(sequence_concat([], [4]), [4])
        self.assertListEqual(sequence_concat([1], [8]), [1, 8])

        self.assertNotIsInstance(sequence_concat([], []), xlist)
        self.assertIsInstance(sequence_concat([], [5]), xlist)
        self.assertNotIsInstance(sequence_concat([1], [8]), xlist)

    def test_count(self):
        self.assertEqual(sequence_count(['a', 'b', 'c']), 3)

    def test_iterate_sequence(self):
        def action(item, pos):
            yield item * pos

        self.assertEqual(iterate_sequence([1, 2, 3], action=action), [1, 4, 9])

        def action(item, pos):
            yield item
            yield item * pos

        self.assertEqual(iterate_sequence([1, 2, 3], action=action), [1, 1, 2, 4, 3, 9])


class XSequenceTest(unittest.TestCase):

    def test_string_repr(self):
        self.assertEqual(repr(XSequence([])), 'XSequence([])')
        self.assertEqual(repr(XSequence([1, 2, 3])), 'XSequence([1, 2, 3])')
        self.assertEqual(repr(XSequence((x for x in [1, 2, 3]))), 'XSequence([1, 2, 3])')
        self.assertEqual(repr(XSequence([1])), 'XSequence([1])')

        self.assertEqual(str(XSequence([1, 2, 3])), '(1, 2, 3)')
        self.assertEqual(str(XSequence([1])), '(1)')
        self.assertEqual(str(XSequence()), '()')

        self.assertEqual(repr(XSequence([1, 2, 3])), 'XSequence([1, 2, 3])')
        self.assertEqual(repr(XSequence([1])), 'XSequence([1])')
        self.assertEqual(repr(XSequence()), 'XSequence([])')

    def test_initialization(self):
        sequence = XSequence([1, 2, 3])
        self.assertEqual(sequence, [1, 2, 3])

    def test_comparison(self):
        self.assertNotEqual(empty_sequence(), ())
        self.assertEqual(empty_sequence(), [])
        self.assertNotEqual((), empty_sequence())
        self.assertEqual([], empty_sequence())
        self.assertNotEqual(empty_sequence(), 0)
        self.assertNotEqual(empty_sequence(), '')
        self.assertNotEqual(empty_sequence(), 1)

        self.assertEqual(XSequence([1]), 1)
        self.assertEqual(XSequence([1]), [1])
        self.assertEqual(XSequence([1]), 1)

    def test_concatenation(self):
        sequence = XSequence([1, 2, 3])
        self.assertEqual(sequence + [4], [1, 2, 3, 4])
        self.assertEqual([] + sequence, [1, 2, 3])
        self.assertEqual([-1, 0] + sequence, [-1, 0, 1, 2, 3])

        s1 = XSequence([1, 2, 3])
        s2 = XSequence([4, 5, 6])
        self.assertEqual(s1 + s2, [1, 2, 3, 4, 5, 6])
        self.assertEqual(s2 + s1, [4, 5, 6, 1, 2, 3])

    def test_sequence_count(self):
        self.assertEqual(XSequence.sequence_count(['a', 'b', 'c']), 3)

    def test_nested_sequence(self):
        self.assertEqual(list(iter(XSequence([]))), [])

        with self.assertRaises(TypeError):
            list(iter(XSequence([[], 8])))
        with self.assertRaises(TypeError):
            list(iter(XSequence([[], [], []])))
        with self.assertRaises(TypeError):
            list(iter(XSequence([[], 8, [9]])))

    def test_empty_sequence(self):
        self.assertEqual(XSequence.empty_sequence(), ())
        self.assertIsInstance(XSequence.empty_sequence(), XSequence)

    def test_sequence_concat(self):
        self.assertEqual(XSequence.sequence_concat([], []), [])
        self.assertIsInstance(XSequence.sequence_concat([], []), XSequence)

        result = XSequence.sequence_concat([90, ], [32])
        self.assertIsInstance(result, XSequence)
        self.assertEqual(empty_sequence(), [])

    def test_iterate_sequence(self):
        def action(item, pos):
            yield item * pos

        result = XSequence.iterate_sequence([1, 2, 3], action=action)
        self.assertEqual(result, [1, 4, 9])
        self.assertIsInstance(result, XSequence)

        def action(item, pos):
            yield item
            yield item * pos

        result = XSequence.iterate_sequence([1, 2, 3], action=action)
        self.assertEqual(result, [1, 1, 2, 4, 3, 9])
        self.assertIsInstance(result, XSequence)


if __name__ == '__main__':
    unittest.main()
